import os
from glob import glob
import numpy as np

_fit_history = []


def _get_file_index():
    os.makedirs("results", exist_ok=True)
    _file_index = 0
    _globbed_files = glob(f"results/fit-{_file_index:03d}-*")
    while len(_globbed_files) > 0:
        _file_index += 1
        _globbed_files = glob(f"results/fit-{_file_index:03d}-*")
    return _file_index


def xy_fit(x_data, y_data, model_function=None, p0=None, dp0=None,
           x_error=None, y_error=None, x_error_rel=None, y_error_rel=None,
           x_error_cor=None, y_error_cor=None, x_error_cor_rel=None, y_error_cor_rel=None,
           errors_rel_to_model=True, limits=None, constraints=None, report=False, profile=None,
           save=True):
    from kafe2.fit.xy.fit import XYFit

    if model_function is None:
        _fit = XYFit([x_data, y_data])
    else:
        _fit = XYFit([x_data, y_data], model_function)
    if p0 is not None:
        _fit.set_all_parameter_values(p0)
    if dp0 is not None:
        _fit.parameter_errors = dp0

    def _add_error_to_fit(axis, error, correlated=False, relative=False):
        if error is None:
            return
        error = np.asarray(error)
        _reference = "model" if errors_rel_to_model and axis == "y" and relative else "data"
        if correlated:
            if error.ndim == 0:
                error = np.reshape(error, (1,))
            for _err in error:
                _fit.add_error(axis, _err, correlation=1.0, reference=_reference)
        else:
            if error.ndim == 2:
                _fit.add_matrix_error(axis, error, "cov", relative=relative, reference=_reference)
            else:
                _fit.add_error(axis, error, relative=relative, reference=_reference)

    _add_error_to_fit("x", x_error)
    _add_error_to_fit("y", y_error)
    _add_error_to_fit("x", x_error_rel, relative=True)
    _add_error_to_fit("y", y_error_rel, relative=True)
    _add_error_to_fit("x", x_error_cor, correlated=True)
    _add_error_to_fit("y", y_error_cor, correlated=True)
    _add_error_to_fit("x", x_error_cor_rel, correlated=True, relative=True)
    _add_error_to_fit("y", y_error_cor_rel, correlated=True, relative=True)

    if limits is not None:
        if not isinstance(limits[0], (list, tuple)):
            limits = (limits,)
        for _limit in limits:
            _fit.limit_parameter(*_limit)
    if constraints is not None:
        if not isinstance(constraints[0], (list, tuple)):
            constraints = (constraints,)
        for _constraint in constraints:
            _fit.add_parameter_constraint(*_constraint)

    if profile is None:
        profile = x_error is not None or x_error_rel is not None or y_error_rel is not None

    _fit_result = _fit.do_fit(asymmetric_parameter_errors=profile)
    if report:
        _fit.report(asymmetric_parameter_errors=profile)

    if save:
        _file_index = _get_file_index()
        _fit.save_state(f"results/fit-{_file_index:03d}-results.yml")
        with open(f"results/fit-{_file_index:03d}-report.txt", "w", encoding="utf8") as _f:
            _fit.report(_f, asymmetric_parameter_errors=profile)
    else:
        _file_index = None

    _fit_history.append(dict(fit=_fit, profile=profile, file_index=_file_index))

    return _fit_result


def plot(fits=-1, x_label=None, y_label=None, data_label=None, model_label=None,
         error_band_label=None, x_range=None, y_range=None, x_scale=None, y_scale=None,
         x_ticks=None, y_ticks=None, parameter_names=None, model_name=None, model_expression=None,
         legend=True, fit_info=True, error_band=True, profile=None, plot_profile=None, show=True,
         save=True):
    from kafe2 import Plot, ContoursProfiler

    _fit_profiles = None
    _file_indices = None
    if isinstance(fits, int):
        if fits >= 0:
            _fit_profiles = [_fit_history[fits]["profile"]]
            _file_indices = [_fit_history[fits]["file_index"]]
            fits = [_fit_history[fits]["fit"]]
        else:
            fits = _fit_history[fits:]
            _fit_profiles = [_f["profile"] for _f in fits]
            _file_indices = [_f["file_index"] for _f in fits]
            fits = [_f["fit"] for _f in fits]
    else:
        try:
            iter(fits)
        except TypeError:
            fits = [fits]
    if parameter_names is not None:
        for _f in fits:
            _f.assign_parameter_latex_names(**parameter_names)
    if model_name is not None:
        if isinstance(model_name, str):
            model_name = [model_name for _ in fits]
        for _f, _mn in zip(fits, model_name):
            _f.assign_model_function_latex_name(_mn)
    if model_expression is not None:
        if isinstance(model_expression, str):
            model_expression = [model_expression for _ in fits]
        for _f, _me in zip(fits, model_expression):
            _f.assign_model_function_latex_expression(_me)
    _plot = Plot(fits)
    if x_label is not None:
        _plot.x_label = x_label
    if y_label is not None:
        _plot.y_label = y_label
    if data_label is not None:
        _plot.customize("data", "label", data_label)
    if model_label is not None:
        _plot.customize("model_line", "label", model_label)
    if error_band_label is not None:
        _plot.customize("model_error_band", "label", error_band_label)
    if not error_band:
        _plot.customize("model_error_band", "label", None)
        _plot.customize("model_error_band", "hide", True)

    if x_range is not None:
        _plot.x_range = x_range
    if y_range is not None:
        _plot.y_range = y_range
    if x_scale is not None:
        _plot.x_scale = x_scale
    if y_scale is not None:
        _plot.y_scale = y_scale
    if x_ticks is not None:
        _plot.x_ticks = x_ticks
    if y_ticks is not None:
        _plot.y_ticks = y_ticks

    if profile is None:
        profile = np.any(_fit_profiles)
    _plot.plot(legend=legend, fit_info=fit_info, asymmetric_parameter_errors=profile)

    _start_index = _get_file_index()
    if save:
        for _i, _ in enumerate(fits):
            _file_index = _start_index + _i if _file_indices is None else _file_indices[_i]
            _plot.save(f"results/fit-{_file_index:03d}-plot.png", dpi=240)

    if plot_profile is None and _fit_profiles is not None:
        plot_profile = _fit_profiles
    if plot_profile is not None:
        try:
            iter(plot_profile)
        except TypeError:
            plot_profile = [plot_profile for _ in fits]
        for _i, (_f_i, _pp_i) in enumerate(zip(fits, plot_profile)):
            if not _pp_i:
                continue
            _cpf = ContoursProfiler(_f_i)
            _cpf.plot_profiles_contours_matrix()
            if save:
                _file_index = _start_index + _i if _file_indices is None else _file_indices[_i]
                _cpf.save(f"results/fit-{_file_index:03d}-profile.png", dpi=240)
    if show:
        _plot.show()


_plot_func = plot


def plot_xy_data(x_data, y_data, x_error=None, y_error=None, x_error_rel=None, y_error_rel=None,
                 x_label=None, y_label=None, data_label=None, x_range=None, y_range=None,
                 x_scale=None, y_scale=None, x_ticks=None, y_ticks=None, show=True, save=True):
    from kafe2.fit.xy.container import XYContainer
    _container = XYContainer(x_data, y_data)
    if x_error is not None:
        _container.add_error("x", x_error)
    if y_error is not None:
        _container.add_error("y", y_error)
    if x_error_rel is not None:
        _container.add_error("x", x_error_rel, relative=True)
    if y_error_rel is not None:
        _container.add_error("y", y_error_rel, relative=True)
    plot(_container, x_label=x_label, y_label=y_label, data_label=data_label, x_range=x_range,
         y_range=y_range, x_scale=x_scale, y_scale=y_scale, x_ticks=x_ticks, y_ticks=y_ticks,
         show=show, save=save)

def k2Fit(func, x, y, sx=None, sy=None, srelx=None, srely=None, xabscor=None, yabscor=None,
          xrelcor=None, yrelcor=None, ref_to_model=True, constraints=None, p0=None, dp0=None,
          limits=None, plot=True, axis_labels=['x-data', 'y-data'], data_legend='data',
          model_expression=None, model_name=None, model_legend='model',
          model_band=r'$\pm 1 \sigma$', fit_info=True, plot_band=True, asym_parerrs=True,
          plot_cor=False, showplots=True, quiet=True):
    xy_fit(
        x, y, func, p0=p0, dp0=dp0, x_error=sx, y_error=sy, x_error_rel=srelx, y_error_rel=srely,
        x_error_cor=xabscor, y_error_cor=yabscor, x_error_cor_rel=xrelcor, y_error_cor_rel=yrelcor,
        errors_rel_to_model=ref_to_model, limits=limits, constraints=constraints, report=not quiet,
        profile=True
    )
    if plot:
        _plot_func(
            x_label=axis_labels[0], y_label=axis_labels[1], data_label=data_legend,
            model_label=model_legend, error_band_label=model_band, model_name=model_name,
            model_expression=model_expression, legend=True, fit_info=fit_info, error_band=plot_band,
            profile=True, plot_profile=plot_cor, show=showplots
        )
    _fit_object = _fit_history[-1]["fit"]
    _parameter_errors = _fit_object.asymmetric_parameter_errors if asym_parerrs else \
        np.stack([-_fit_object.parameter_errors, _fit_object.parameter_errors], axis=-1)
    return (_fit_object.parameter_values, _parameter_errors, _fit_object.parameter_cor_mat,
            _fit_object.goodness_of_fit)



import abc
import numpy as np
import six

from ..fit.io.file import FileIOMixin

__all__ = ['ParameterConstraint', 'GaussianSimpleParameterConstraint',
           'GaussianMatrixParameterConstraint']


@six.add_metaclass(abc.ABCMeta)
class ParameterConstraint(FileIOMixin, object):
    """
    Abstract base class for parameter constraints.
    Subclasses must implement the ``cost`` method.
    """

    def __init__(self):
        pass

    @property
    @abc.abstractmethod
    def extra_ndf(self):
        """
        :return: the additional number of degrees of freedom introduced by this constraint.
        """

    def _get_base_class(self):
        return ParameterConstraint

    def _get_object_type_name(self):
        return 'parameter_constraint'

    def cost(self, parameter_values):
        """
        Calculates additional cost depending on the fit parameter values.

        :param parameter_values: The current parameter values of the fit
        :type parameter_values: iterable of float
        :return: The additional cost imposed by the given parameter values
        :rtype: float
        """
        pass


class GaussianSimpleParameterConstraint(ParameterConstraint):

    def __init__(self, index, value, uncertainty, relative=False):
        """
        Simple class for applying a gaussian constraint to a single parameter of a fit.

        :param index: The index of the parameter to be constrained
        :type index: int
        :param value: The value to which the parameter should be constrained
        :type value: float
        :param uncertainty: The uncertainty with which the parameter should be constrained to the given value
        :type uncertainty: float
        :param relative: Whether the given uncertainty is relative to the given value
        :type relative: bool
        """
        self._index = index
        self._value = value
        if relative:
            self._uncertainty_abs = None
            self._uncertainty_rel = uncertainty
        else:
            self._uncertainty_abs = uncertainty
            self._uncertainty_rel = None
        self._relative = relative
        super(GaussianSimpleParameterConstraint, self).__init__()

    @property
    def index(self):
        """the index of the constrained parameter"""
        return self._index

    @property
    def value(self):
        """the value to which the parameter is being constrained"""
        return self._value

    @property
    def uncertainty(self):
        """the absolute uncertainty with which the parameter is being constrained"""
        if self._uncertainty_abs is None:
            self._uncertainty_abs = self._uncertainty_rel * self.value
        return self._uncertainty_abs

    @property
    def uncertainty_rel(self):
        """the uncertainty relative to ``value`` with which the parameter is being constrained"""
        if self._uncertainty_rel is None:
            self._uncertainty_rel = self._uncertainty_abs / self.value
        return self._uncertainty_rel

    @property
    def relative(self):
        """whether the constraint was initialized with a relative uncertainty"""
        return self._relative

    @property
    def extra_ndf(self):
        return 1

    def cost(self, parameter_values):
        """
        Calculates additional cost depending on the fit parameter values.
        More specifically, the constraint first picks the value from ``parameter_values`` at ``self.index``.
        The constraint then calculates the residual by subtracting ``self.value``.
        The final cost is calculated by dividing the residual by ``self.uncertainty`` and squaring the result.

        :param parameter_values: The current parameter values of the fit
        :type parameter_values: iterable of float
        :return: The additional cost imposed by the given parameter values
        :rtype: float
        """
        return ((parameter_values[self.index] - self.value) / self.uncertainty) ** 2


class GaussianMatrixParameterConstraint(ParameterConstraint):

    def __init__(self, indices, values, matrix, matrix_type='cov', uncertainties=None, relative=False):
        """
        Advanced class for applying correlated constraints to several parameters of a fit.
        The order of ``indices``, ``values``, ``matrix``, and ``uncertainties`` must be aligned.
        In other words the first index must belong to the first value, the first row/column in the matrix, etc.

        Let N be the number of parameters to be constrained.
        :param indices: The indices of the parameters to be constrained
        :type indices: iterable of int, shape (N,)
        :param values: The values to which the parameters should be constrained
        :type values: iterable of float, shape (N,)
        :param matrix: The matrix that defines the correlation between the parameters. By default interpreted as a
            covariance matrix. Can also be interpreted as a correlation matrix by setting ``matrix_type``
        :type matrix: iterable of float, shape (N, N)
        :param matrix_type: Whether the matrix should be interpreted as a covariance matrix or as a correlation matrix
        :type matrix_type: str, either 'cov' or 'cor'
        :param uncertainties: The uncertainties to be used in conjunction with a correlation matrix
        :type uncertainties: ``None`` or iterable of float, shape (N,)
        :param relative: Whether the covariance matrix/the uncertainties should be interpreted as relative to ``values``
        :type relative: bool
        """
        self._indices = np.array(indices)
        self._values = np.array(values)

        _matrix_array = np.array(matrix)
        if not np.array_equal(_matrix_array, _matrix_array.T):
            raise ValueError('The matrix for parameter constraints must be symmetric!')
        if len(self._values.shape) != 1 or self._values.shape * 2 != _matrix_array.shape:
            raise ValueError(
                'Expected values and cov_mat to be of shapes (N, ), (N, N) but received shapes %s, %s instead!'
                % (self._values.shape, _matrix_array.shape))
        if matrix_type == 'cov':
            pass
        elif matrix_type == 'cor':
            if np.any(np.diag(_matrix_array) != 1.0):
                raise ValueError('The correlation matrix has diagonal elements that aren\'t equal to 1!')
            if np.any(_matrix_array> 1.0):
                raise ValueError('The correlation matrix has elements greater than 1!')
            if np.any(_matrix_array < -1.0):
                raise ValueError('The correlation matrix has elements smaller than -1!')
        else:
            raise ValueError('Unknown matrix_type: %s, must be either cov or cor!' % matrix_type)

        if matrix_type == 'cov':
            if relative:
                self._cov_mat_abs = None
                self._cov_mat_rel = _matrix_array
                self._cor_mat = None
            else:
                self._cov_mat_abs = _matrix_array
                self._cov_mat_rel = None
                self._cor_mat = None

            if uncertainties is not None:
                raise ValueError('Uncertainties can only be specified if matrix_type is cov!')
            self._uncertainties_abs = None
            self._uncertainties_rel = None
        else:
            self._cov_mat_abs = None
            self._cov_mat_rel = None
            self._cor_mat = _matrix_array
            if uncertainties is None:
                raise ValueError('If matrix_type is cor uncertainties must be specified!')
            if relative:
                self._uncertainties_abs = None
                self._uncertainties_rel = uncertainties
            else:
                self._uncertainties_abs = uncertainties
                self._uncertainties_rel = None

        self._matrix_type = matrix_type
        self._relative = relative

        self._cov_mat_inverse = None
        super(GaussianMatrixParameterConstraint, self).__init__()

    @property
    def indices(self):
        """the indices of the parameters to be constrained"""
        return self._indices

    @property
    def values(self):
        """the values to which the parameters are being constrained"""
        return self._values

    @property
    def cov_mat(self):
        """the absolute covariance matrix between the parameter uncertainties"""
        if self._cov_mat_abs is None:
            if self.matrix_type == 'cov':
                self._cov_mat_abs = self._cov_mat_rel * np.outer(self.values, self.values)
            else:
                self._cov_mat_abs = self._cor_mat * np.outer(self.uncertainties, self.uncertainties)
        return self._cov_mat_abs

    @property
    def cov_mat_rel(self):
        """the covariance matrix between the parameter uncertainties relative to ``self.values``"""
        if self._cov_mat_rel is None:
            if self.matrix_type == 'cov':
                self._cov_mat_rel = self._cov_mat_abs / np.outer(self.values, self.values)
            else:
                self._cov_mat_rel = self._cor_mat * np.outer(self.uncertainties_rel, self.uncertainties_rel)
        return self._cov_mat_rel

    @property
    def cor_mat(self):
        """the correlation matrix between the parameter uncertainties"""
        if self._cor_mat is None:
            # if the originally specified cov mat was relative, calculate the cor mat based on that
            if self._relative:
                self._cor_mat = self.cov_mat_rel / np.outer(self.uncertainties_rel, self.uncertainties_rel)
            else:
                self._cor_mat = self.cov_mat / np.outer(self.uncertainties, self.uncertainties)
        return self._cor_mat

    @property
    def uncertainties(self):
        """the uncorrelated, absolute uncertainties for the parameters to be constrained to"""
        if self._uncertainties_abs is None:
            if self.matrix_type == 'cov':
                self._uncertainties_abs = np.sqrt(np.diag(self.cov_mat))
            else:
                self._uncertainties_abs = self.uncertainties_rel * self.values
        return self._uncertainties_abs

    @property
    def uncertainties_rel(self):
        """the uncorrelated uncertainties for the parameters to be constrained to relative to ``self.values``"""
        if self._uncertainties_rel is None:
            if self.matrix_type == 'cov':
                self._uncertainties_rel = np.sqrt(np.diag(self.cov_mat_rel))
            else:
                self._uncertainties_rel = self.uncertainties / self.values
        return self._uncertainties_rel

    @property
    def matrix_type(self):
        """the type of matrix with which the constraint was initialized"""
        return self._matrix_type

    @property
    def relative(self):
        """whether the constraint was initialized with a relative covariance matrix/with relative uncertainties"""
        return self._relative

    @property
    def cov_mat_inverse(self):
        """the inverse of the covariance matrix between the parameter uncertainties"""
        if self._cov_mat_inverse is None:
            self._cov_mat_inverse = np.linalg.inv(self.cov_mat)
        return self._cov_mat_inverse

    @property
    def extra_ndf(self):
        return len(self.indices)

    def cost(self, parameter_values):
        """
        Calculates additional cost depending on the fit parameter values.
        More specifically, the constraint first picks values from ``parameter_values`` according to ``self.indices``.
        The constraint then calculates the residuals by subtracting ``self.values``.
        The final cost is calculated by applying the residuals to both sides of ``self.cov_mat_inverse``
        via dot product.

        :param parameter_values: The current parameter values of the fit
        :type parameter_values: iterable of float
        :return: The additional cost imposed by the given parameter values
        :rtype: float
        """
        _selected_par_values = np.asarray(parameter_values)[self.indices]
        _res = _selected_par_values - self.values
        return _res.dot(self.cov_mat_inverse).dot(_res)



import unittest
import numpy as np

from kafe2.fit import CustomFit

from kafe2.test.fit.test_fit import AbstractTestFit

class TestCustomFitWithSimpleYErrors(AbstractTestFit, unittest.TestCase):

    MINIMIZER = 'scipy'

    @staticmethod
    def chi2(a=1.5, b=-0.5):
        x_data = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0])
        y_data = np.array([
            -1.0804945, 0.97336504, 2.75769933, 4.91093935, 6.98511206, 9.15059627, 10.9665515,
            13.06741151, 14.95081026, 16.94404467
        ])
        y_err = 0.1
        y_model = a * x_data + b
        cost = np.sum(np.square((y_model - y_data) / y_err))
        cost += 2 * x_data.shape[0] * np.log(y_err)
        return cost

    def setUp(self):
        self._n_points = 10

        # reference initial values
        self._ref_initial_pars = np.array([1.5, -0.5])

        # pre-fit cost value
        self._ref_initial_cost = self.chi2(*self._ref_initial_pars)

        self._nominal_fit_result_pars = np.array([2.0117809, -1.09041046])

        self._nominal_fit_result_cost = self.chi2(*self._nominal_fit_result_pars)

        # helper dict with all reference property values
        self._ref_prop_dict = dict(
            did_fit=False,
            model_count=0,

            parameter_values=self._ref_initial_pars,
            parameter_names=('a', 'b'),
            cost_function_value=self._ref_initial_cost,

            data=None,
            data_error=None,
            data_cov_mat=None,
            data_cov_mat_inverse=None,
            data_cor_mat=None,

            model=None,
            model_error=None,
            model_cov_mat=None,
            model_cov_mat_inverse=None,
            model_cor_mat=None,

            total_error=None,
            total_cov_mat=None,
            total_cov_mat_inverse=None,
            total_cor_mat=None,
        )

    def _get_fit(self):
        '''convenience'''
        _fit = CustomFit(
            cost_function=self.chi2,
            minimizer=self.MINIMIZER
        )
        return _fit

    def _get_test_fits(self):
        return {
            'default': self._get_fit(),
        }

    def test_initial_state(self):
        self.run_test_for_all_fits(
            self._ref_prop_dict,
            fit_names=['default'],
        )

    def test_fit_results(self):
        self.run_test_for_all_fits(
            dict(self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                did_fit=True,
                cost_function_value=self._nominal_fit_result_cost,
            ),
            fit_names=['default'],
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-2
        )



import inspect
import numpy as np

from ._base import FitDReprBase
from .. import _AVAILABLE_REPRESENTATIONS
from .._yaml_base import YamlWriterMixin, YamlReaderMixin
from ..constraint.yaml_drepr import ConstraintYamlReader, ConstraintYamlWriter
from ..container.yaml_drepr import DataContainerYamlReader, DataContainerYamlWriter
from ..model.yaml_drepr import ParametricModelYamlReader, ParametricModelYamlWriter, \
    _process_function_code_for_dump, _parse_function
from ....fit import CustomFit, HistFit, IndexedFit, UnbinnedFit, XYFit
from ..._base.cost import STRING_TO_COST_FUNCTION
from ..._base.format import ParameterFormatter
from ...xy.cost import STRING_TO_COST_FUNCTION as STRING_TO_COST_FUNCTION_XY
from ...util import to_python_types, to_numpy_arrays
from ...unbinned.cost import STRING_TO_COST_FUNCTION as STRING_TO_COST_FUNCTION_UNBINNED
from ....tools import get_compact_representation

__all__ = ['FitYamlWriter', 'FitYamlReader']


class FitYamlWriter(YamlWriterMixin, FitDReprBase):

    def __init__(self, fit, output_io_handle):
        super(FitYamlWriter, self).__init__(
            output_io_handle=output_io_handle,
            fit=fit)

    def _get_preface_comment(self):
        _preface_comment = super(FitYamlWriter, self)._get_preface_comment()
        _did_fit = self._kafe_object.did_fit
        if not _did_fit:
            _preface_comment += "\n# WARNING: No fit has been performed as of yet. " \
                                "Did you forget to run fit.do_fit()?\n"
        _preface_comment += '\n'
        if self._kafe_object.model_count == 1:
            _preface_comment += "# Model function: %s\n" % self._kafe_object._model_function.formatter.get_formatted(
                format_as_latex=False,
                with_expression=True,
                with_par_values=False
            )
        else:
            for _i in range(self._kafe_object.model_count):
                _preface_comment += "# Model function %s: %s\n" % (_i,
                    self._kafe_object._model_function.formatter.get_formatted(
                        model_index=_i,
                        format_as_latex=False,
                        with_expression=True,
                        with_par_values=False
                    )
                )

        if _did_fit:
            _gof = self._kafe_object.goodness_of_fit
            _gof_name = "chi2" if self._kafe_object._cost_function.is_chi2 else "GoF"
            _cost = self._kafe_object.cost_function_value
            _ndf = self._kafe_object.ndf
            if _gof is None:
                _round_cost_sig = max(2, int(-np.floor(np.log(np.abs(_cost)) / np.log(10))) + 1)
                _rounded_cost = round(_cost, _round_cost_sig)
                _preface_comment += "# Cost: %s\n" % _rounded_cost
            else:
                _preface_comment += "# %s: %s\n" % (_gof_name, _gof)
                _round_gof_per_ndf_sig = max(
                    2, int(-np.floor(np.log(np.abs(_gof) / _ndf) / np.log(10))) + 1)
            if _ndf is not None:
                _preface_comment += '# ndf: %s\n' % _ndf
            if _gof is not None:
                _preface_comment += "# %s/ndf: %s\n\n" % (
                    _gof_name, round(_gof / _ndf, _round_gof_per_ndf_sig))

            # If asymmetric parameters errors were not calculated, check the loaded result dict
            _asymmetric_parameter_errors = self._kafe_object._fitter.asymmetric_fit_parameter_errors_if_calculated
            if _asymmetric_parameter_errors is None and self._kafe_object._loaded_result_dict is not None:
                _asymmetric_parameter_errors = self._kafe_object._loaded_result_dict['asymmetric_parameter_errors']

            _preface_comment += get_compact_representation(
                parameter_names=self._kafe_object.parameter_names,
                parameter_values=self._kafe_object.parameter_values,
                parameter_errors=self._kafe_object.parameter_errors,
                parameter_cor_mat=self._kafe_object.parameter_cor_mat,
                asymmetric_parameter_errors=_asymmetric_parameter_errors
            )
        _preface_comment += '\n'
        return _preface_comment

    @classmethod
    def _make_representation(cls, fit):
        _yaml_doc = dict()

        # -- determine fit type
        _type = cls._CLASS_TO_OBJECT_TYPE_NAME.get(fit.__class__, None)
        if _type is None:
            raise TypeError("Fit type unknown or not supported: %s" % fit.__class__)
        _yaml_doc['type'] = _type

        if _type != 'custom':
            _yaml_doc['dataset'] = DataContainerYamlWriter._make_representation(
                fit.data_container)
            _yaml_doc['parametric_model'] = ParametricModelYamlWriter._make_representation(
                fit._param_model)
        else:
            _par_formatter_dict = {
                _par_formatter.arg_name: _par_formatter.name
                for _par_formatter in fit._parameter_formatters
            }
            if _par_formatter_dict:
                _yaml_doc['parameter_formatters'] = _par_formatter_dict

        _cost_function_identifier = fit._cost_function.kafe2go_identifier
        if _cost_function_identifier is not None:
            _yaml_doc['cost_function'] = _cost_function_identifier
        else:
            _yaml_doc['cost_function'] = _process_function_code_for_dump(
                inspect.getsource(fit._cost_function.func))

        _yaml_doc['minimizer'] = fit._minimizer
        _yaml_doc['minimizer_kwargs'] = fit._minimizer_kwargs

        _yaml_doc['parameter_constraints'] = [ConstraintYamlWriter._make_representation(_parameter_constraint)
                                              for _parameter_constraint in fit.parameter_constraints]
        _yaml_doc['fixed_parameters'] = fit._fitter.fixed_parameters
        _yaml_doc['limited_parameters'] = {
            _par_name: list(_par_limits)
            for _par_name, _par_limits in fit._fitter.limited_parameters.items()}

        _yaml_doc['fit_results'] = to_python_types(fit.get_result_dict())
        return _yaml_doc


class FitYamlReader(YamlReaderMixin, FitDReprBase):

    def __init__(self, input_io_handle):
        super(FitYamlReader, self).__init__(
            input_io_handle=input_io_handle,
            fit=None)

    @classmethod
    def _get_required_keywords(cls, yaml_doc, fit_class):
        if fit_class is CustomFit:
            return ['cost_function']
        if fit_class in (HistFit, XYFit):
            return ['dataset']
        return ['dataset', 'parametric_model']

    @classmethod
    def _get_subspace_override_dict(cls, fit_class):
        _override_dict = {'model_parameters': 'parametric_model',
                          'arg_formatters': 'parametric_model',
                          'model_function_formatter': 'parametric_model',
                          'expression_string': 'parametric_model',
                          'latex_expression_string': 'parametric_model',
                          }
        if fit_class is CustomFit:
            pass
        elif fit_class is HistFit:
            _override_dict['n_bins'] = ['dataset', 'parametric_model']
            _override_dict['bin_range'] = ['dataset', 'parametric_model']
            _override_dict['bin_edges'] = ['dataset', 'parametric_model']
            _override_dict['raw_data'] = 'dataset'
            _override_dict['errors'] = 'dataset'
            _override_dict['model_density_function'] = 'parametric_model'
            _override_dict['model_density_function_name'] = 'parametric_model'
            _override_dict['latex_model_density_function_name'] = 'parametric_model'
        elif fit_class is IndexedFit:
            _override_dict['data'] = 'dataset'
            _override_dict['errors'] = 'dataset'
            _override_dict['model_function'] = 'parametric_model'
            _override_dict['model_function_name'] = 'parametric_model'
            _override_dict['latex_model_function_name'] = 'parametric_model'
            _override_dict['index_name'] = 'parametric_model'
            _override_dict['latex_index_name'] = 'parametric_model'
        elif fit_class is UnbinnedFit:
            _override_dict['data'] = ['dataset', 'parametric_model']
            _override_dict['model_function'] = 'parametric_model'
            _override_dict['model_function_name'] = 'parametric_model'
            _override_dict['latex_model_function_name'] = 'parametric_model'
        elif fit_class is XYFit:
            _override_dict['x_data'] = ['dataset', 'parametric_model']
            _override_dict['y_data'] = 'dataset'
            _override_dict['x_errors'] = 'dataset'
            _override_dict['y_errors'] = 'dataset'
            _override_dict['model_function'] = 'parametric_model'
            _override_dict['model_function_name'] = 'parametric_model'
            _override_dict['latex_model_function_name'] = 'parametric_model'
        else:
            raise TypeError("Unknown fit type")
        # override labels for every fit type
        _override_dict['label'] = 'dataset'
        _override_dict['model_label'] = 'parametric_model'
        _override_dict['x_label'] = 'dataset'
        _override_dict['y_label'] = 'dataset'
        return _override_dict

    @classmethod
    def _convert_yaml_doc_to_object(cls, yaml_doc):
        # -- determine fit class from type
        _fit_type = yaml_doc.pop('type')
        _class = cls._OBJECT_TYPE_NAME_TO_CLASS.get(_fit_type, None)

        if _fit_type != 'custom':
            _data = DataContainerYamlReader._make_object(
                yaml_doc.pop('dataset'), default_type=_fit_type)
        _parametric_model_entry = yaml_doc.pop('parametric_model', None)
        if _parametric_model_entry:
            _read_parametric_model = ParametricModelYamlReader._make_object(
                _parametric_model_entry, default_type=_fit_type, dataset=_data)
            _read_model_function = _read_parametric_model._model_function_object
        else:
            _read_parametric_model = None
            _read_model_function = None

        _cost_function = yaml_doc.pop('cost_function', None)
        if _cost_function is not None:
            if _fit_type == 'xy':
                _lookup_dict = STRING_TO_COST_FUNCTION_XY
            elif _fit_type == 'unbinned':
                _lookup_dict = STRING_TO_COST_FUNCTION_UNBINNED
            else:
                _lookup_dict = STRING_TO_COST_FUNCTION
            if _cost_function not in _lookup_dict:
                _cost_function = _parse_function(_cost_function)

        _minimizer = yaml_doc.pop('minimizer', None)
        _minimizer_kwargs = yaml_doc.pop('minimizer_kwargs', None)
        # change fit kwargs for different fit types if necessary
        _fit_kwargs = dict(minimizer=_minimizer, minimizer_kwargs=_minimizer_kwargs)
        if _cost_function is not None:
            _fit_kwargs["cost_function"] = _cost_function
        if _fit_type != 'custom':
            _fit_object = _class(_data, _read_model_function, **_fit_kwargs)
        else:
            _fit_object = _class(**_fit_kwargs)
            _par_formatters = yaml_doc.pop('parameter_formatters', None)
            if _par_formatters is not None:
                _fit_object._parameter_formatters = [
                    ParameterFormatter(arg_name=_arg_name, name=_name)
                    for (_arg_name, _name) in _par_formatters.items()
                ]
            _fit_object._update_parameter_formatters()

        if _read_parametric_model is not None:
            _fit_object._param_model = _read_parametric_model

        _constraint_yaml_list = yaml_doc.pop('parameter_constraints', None)
        if isinstance(_constraint_yaml_list, dict):
            _constraint_yaml_list = [dict(name=param, **_constraint_yaml)
                                     for param, _constraint_yaml in _constraint_yaml_list.items()]
        if isinstance(_constraint_yaml_list, list):
            _fit_object._fit_param_constraints = [
                ConstraintYamlReader._make_object(
                    _constraint_yaml, default_type='simple',
                    parameter_names=_fit_object.parameter_names)
                for _constraint_yaml in _constraint_yaml_list
            ]

        _fixed_par_list = yaml_doc.pop('fixed_parameters', None)
        if _fixed_par_list is not None:
            for _par, _val in _fixed_par_list.items():
                _fit_object.fix_parameter(_par, _val)

        _limited_par_list = yaml_doc.pop('limited_parameters', None)
        if _limited_par_list is not None:
            for _par, _limits in _limited_par_list.items():
                _low, _high = _limits
                _fit_object.limit_parameter(_par, _low, _high)

        _fit_results = yaml_doc.pop('fit_results', None)
        _fit_object._loaded_result_dict = to_numpy_arrays(_fit_results)
        return _fit_object, yaml_doc


# register the above classes in the module-level dictionary
FitYamlReader._register_class(_AVAILABLE_REPRESENTATIONS)
FitYamlWriter._register_class(_AVAILABLE_REPRESENTATIONS)



#!/usr/bin/env python3

import unittest
import numpy as np

from kafe2.fit.indexed.model import IndexedModelFunction
from kafe2.fit.histogram.model import HistModelFunction
from kafe2.fit._base.model import ModelFunctionBase


class TestSymPy(unittest.TestCase):
    MODEL_FUNCTION_CLASS = ModelFunctionBase

    def setUp(self):

        def model(x, a, b, c):
            return a * x ** 2 + b * x + c

        self._model_1_python = model
        self._model_1_sympy = "x a b c -> a * x^2 + b * x + c"

        self._ref_properties_1 = dict(
            x=np.arange(10),
            parameter_values=[1.2, 3.4, 5.6],
            name="model",
            parameter_names=["a", "b", "c"],
            argcount=4,
            parcount=3,
            x_name="x",
            defaults=[1.0, 1.0, 1.0],
            latex_name=r"{\tt model}",
            parameter_latex_names=["{a}", "{b}", "{c}"]
        )
        _x = self._ref_properties_1["x"]
        _pvs = self._ref_properties_1["parameter_values"]
        self._ref_properties_1["y"] = _pvs[0] * _x ** 2 + _pvs[1] * _x + _pvs[2]

        def other_model(tau, A_0=3.0, tau_0=1.0):
            return A_0 * np.exp(-tau / tau_0)

        self._model_2_python = other_model
        self._model_2_sympy = "other_model: tau A_0=3.0 tau_0 -> A_0 * exp(-tau / tau_0)"

        self._ref_properties_2 = dict(
            x=np.arange(10),
            parameter_values=[1.2, 3.4],
            name="other_model",
            parameter_names=["A_0", "tau_0"],
            argcount=3,
            parcount=2,
            x_name="tau",
            defaults=[3.0, 1.0],
            latex_name=r"{\tt other\_model}",
            parameter_latex_names=["{A_0}", r"{\tau_0}"]
        )
        _x = self._ref_properties_2["x"]
        _pvs = self._ref_properties_2["parameter_values"]
        self._ref_properties_2["y"] = _pvs[0] * np.exp(-_x / _pvs[1])

    def _assert_properties_correct(self, model_function, ref_properties):
        with self.subTest("y"):
            self.assertTrue(np.allclose(
                model_function(ref_properties["x"], *ref_properties["parameter_values"]),
                ref_properties["y"]
            ))
        with self.subTest("name"):
            self.assertEqual(model_function.name, ref_properties["name"])
        with self.subTest("parameter_names"):
            self.assertEqual(model_function.parameter_names, ref_properties["parameter_names"])
        with self.subTest("argcount"):
            self.assertEqual(model_function.argcount, ref_properties["argcount"])
        with self.subTest("parcount"):
            self.assertEqual(model_function.parcount, ref_properties["parcount"])
        with self.subTest("x_name"):
            self.assertEqual(model_function.x_name, ref_properties["x_name"])
        with self.subTest("defaults"):
            self.assertEqual(model_function.defaults, ref_properties["defaults"])
        with self.subTest("formatter_name"):
            self.assertEqual(model_function.formatter.name, ref_properties["name"])
        with self.subTest("formatter_latex_name"):
            self.assertEqual(model_function.formatter.latex_name, ref_properties["latex_name"])
        with self.subTest("formatter_parameter_names"):
            self.assertEqual(
                [_pf.name for _pf in model_function.formatter.par_formatters],
                ref_properties["parameter_names"]
            )
        with self.subTest("formatter_parameter_latex_names"):
            self.assertEqual(
                [_pf.latex_name for _pf in model_function.formatter.par_formatters],
                ref_properties["parameter_latex_names"]
            )

    def test_properties_python_1(self):
        model_function = self.MODEL_FUNCTION_CLASS(self._model_1_python)
        self._assert_properties_correct(model_function, self._ref_properties_1)

    def test_properties_sympy_1(self):
        model_function = self.MODEL_FUNCTION_CLASS(self._model_1_sympy)
        self._assert_properties_correct(model_function, self._ref_properties_1)

    def test_properties_python_2(self):
        model_function = self.MODEL_FUNCTION_CLASS(self._model_2_python)
        self._assert_properties_correct(model_function, self._ref_properties_2)

    def test_properties_sympy_2(self):
        model_function = self.MODEL_FUNCTION_CLASS(self._model_2_sympy)
        self._assert_properties_correct(model_function, self._ref_properties_2)


class TestSymPyHist(TestSymPy):
    MODEL_FUNCTION_CLASS = HistModelFunction


class TestSymPyIndexed(TestSymPy):
    MODEL_FUNCTION_CLASS = IndexedModelFunction

    def setUp(self):
        super().setUp()
        self._ref_properties_1["parameter_names"].insert(0, self._ref_properties_1["x_name"])
        self._ref_properties_1["parcount"] += 1
        self._ref_properties_1["x_name"] = None
        self._ref_properties_1["defaults"].insert(0, 1.0)
        self._ref_properties_1["parameter_latex_names"].insert(0, r"{x}")

        self._ref_properties_2["parameter_names"].insert(0, self._ref_properties_2["x_name"])
        self._ref_properties_2["parcount"] += 1
        self._ref_properties_2["x_name"] = None
        self._ref_properties_2["defaults"].insert(0, 1.0)
        self._ref_properties_2["parameter_latex_names"].insert(0, r"{\tau}")



"""This submodule contains tools for writing and reading to streams and files.

:synopsis: This submodule contains tools for writing and reading to streams and files.

.. moduleauthor:: Daniel Savoiu <daniel.savoiu@cern.ch>
"""

from .handle import *



import unittest

from kafe2.core.confidence import ConfidenceLevel


class TestConfidence(unittest.TestCase):

    def setUp(self):
        self._sigma = [1.0, 2.0, 3.0]
        self._cl_1d = [0.6827, 0.9545, 0.9973]
        self._cl_2d = [0.3935, 0.8647, 0.9889]

    def test_one_way_conversion(self):
        for _sigma, _cl_1d in zip(self._sigma, self._cl_1d):
            self.assertAlmostEqual(
                _cl_1d,
                ConfidenceLevel(n_dimensions=1, sigma=_sigma).cl,
                places=4
            )
            self.assertAlmostEqual(
                _sigma,
                ConfidenceLevel(n_dimensions=1, cl=_cl_1d).sigma,
                places=3
            )
        for _sigma, _cl_2d in zip(self._sigma, self._cl_2d):
            self.assertAlmostEqual(
                _cl_2d,
                ConfidenceLevel(n_dimensions=2, sigma=_sigma).cl,
                places=4
            )
            self.assertAlmostEqual(
                _sigma,
                ConfidenceLevel(n_dimensions=2, cl=_cl_2d).sigma,
                places=3
            )

    def test_round_trip(self):
        for _n_dimensions in range(1, 10):
            for _sigma in [0.5, 1.2, 2.3, 3.4, 4.5]:
                self.assertAlmostEqual(
                    _sigma,
                    ConfidenceLevel(
                        n_dimensions=_n_dimensions,
                        cl=ConfidenceLevel(
                            n_dimensions=_n_dimensions,
                            sigma=_sigma
                        ).cl
                    ).sigma
                )
            for _cl in [0.12, 0.34, 0.56, 0.78, 0.99]:
                self.assertAlmostEqual(
                    _cl,
                    ConfidenceLevel(
                        n_dimensions=_n_dimensions,
                        sigma=ConfidenceLevel(
                            n_dimensions=_n_dimensions,
                            cl=_cl
                        ).sigma
                    ).cl
                )

    def test_constructor_raise(self):
        with self.assertRaises(ValueError):
            ConfidenceLevel(n_dimensions=1)
        with self.assertRaises(ValueError):
            ConfidenceLevel(n_dimensions=1, cl=0.683, sigma=1.0)
        with self.assertRaises(ValueError):
            ConfidenceLevel(n_dimensions=1, cl=-1.0)
        with self.assertRaises(ValueError):
            ConfidenceLevel(n_dimensions=1, cl=2.0)
        with self.assertRaises(ValueError):
            ConfidenceLevel(n_dimensions=1, sigma=-1)
        with self.assertRaises(ValueError):
            ConfidenceLevel(n_dimensions=1.1, sigma=1)
        with self.assertRaises(ValueError):
            ConfidenceLevel(n_dimensions="1", sigma=1)
        with self.assertRaises(ValueError):
            ConfidenceLevel(n_dimensions=0, sigma=1)
        with self.assertRaises(ValueError):
            ConfidenceLevel(n_dimensions=-1, sigma=1)

    def test_string_represenations(self):
        _cl = ConfidenceLevel(n_dimensions=1, sigma=1)
        self.assertEqual(str(_cl), "<ConfidenceLevel (d=1): 68.27% (1-sigma)>")
        self.assertEqual(_cl.sigma_string, "1-sigma")
        self.assertEqual(_cl.sigma_latex_string, "1$\\sigma$")
        self.assertEqual(_cl.cl_string, "68.27% CL")
        self.assertEqual(_cl.cl_latex_string, "$68.27\\%$ CL")



import abc
import six

from .._base import GenericDReprBase
# import data container classes
from ...histogram import HistContainer
from ...indexed import IndexedContainer
from ...unbinned import UnbinnedContainer
from ...xy import XYContainer

__all__ = ["DataContainerDReprBase"]


@six.add_metaclass(abc.ABCMeta)
class DataContainerDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = 'container'

    _CLASS_TO_OBJECT_TYPE_NAME = {
        HistContainer: 'histogram',
        IndexedContainer: 'indexed',
        UnbinnedContainer: 'unbinned',
        XYContainer: 'xy'
    }
    _OBJECT_TYPE_NAME_TO_CLASS = {
        'histogram': HistContainer,
        'indexed': IndexedContainer,
        'unbinned': UnbinnedContainer,
        'xy': XYContainer
    }

    def __init__(self, data_container=None):
        self._kafe_object = data_container
        super(DataContainerDReprBase, self).__init__()



import unittest
import numpy as np
from six import StringIO

from kafe2.fit._base import ModelFunctionBase
from kafe2.fit.histogram import HistModelFunction
from kafe2.fit.indexed import IndexedModelFunction
from kafe2.fit.representation import ParametricModelYamlWriter, ParametricModelYamlReader
from kafe2.fit.histogram.model import HistParametricModel
from kafe2.fit.indexed.model import IndexedParametricModel
from kafe2.fit.xy.model import XYParametricModel
from kafe2.fit.io.handle import IOStreamHandle

TEST_PARAMETRIC_MODEL_HIST="""
type: histogram
n_bins: 5
bin_range: [0, 5]
model_density_function:
    type: histogram
    python_code: |
        def linear_model(x, a, b):
            return a * x + b
model_parameters: [0.0, 0.08]
"""

TEST_PARAMETRIC_MODEL_HIST_MISSING_KEYWORD="""
type: histogram
n_bins: 5
bin_range: [0, 5]
model_parameters: [0.0, 0.08]
"""

TEST_PARAMETRIC_MODEL_HIST_EXTRA_KEYWORD = TEST_PARAMETRIC_MODEL_HIST + """
extra_keyword: 3.14
"""

TEST_PARAMETRIC_MODEL_HIST_WITH_ERRORS = TEST_PARAMETRIC_MODEL_HIST + """
errors:
  - correlation_coefficient: 0.0
    error_value: 0.1
    name: test_error
    relative: false
    type: simple
"""

TEST_PARAMETRIC_MODEL_HIST_LEGACY = TEST_PARAMETRIC_MODEL_HIST + """
model_density_func_antiderivative: null
"""

TEST_PARAMETRIC_MODEL_HIST_ANTIDERIVATIVE = TEST_PARAMETRIC_MODEL_HIST + """
bin_evaluation: |
    def linear_model_antiderivative(x, a, b):
        return 0.5 * a * x ** 2 + b * x
"""

TEST_PARAMETRIC_MODEL_HIST_NUMPY_VECTORIZE = TEST_PARAMETRIC_MODEL_HIST + """
bin_evaluation: |
    @np.vectorize
    def linear_model_antiderivative(x, a, b):
        return 0.5 * a * x ** 2 + b * x
"""


class TestHistParametricModelYamlRepresenter(unittest.TestCase):

    def _assert_read_model_correct(self, read_model, ref_model, errors=False):
        self.assertTrue(isinstance(read_model, HistParametricModel))

        self.assertTrue(np.all(
            read_model.data == ref_model.data
        ))
        if isinstance(ref_model.bin_evaluation, str):
            self.assertEqual(read_model.bin_evaluation, ref_model.bin_evaluation)
            self.assertEqual(read_model.bin_evaluation_string, ref_model.bin_evaluation_string)
        self.assertTrue(np.all(
            read_model.parameters == self._test_model_parameters
        ))
        self.assertTrue(read_model.n_bins == self._test_n_bins)
        self.assertTrue(np.all(
            read_model.bin_range == self._test_bin_range
        ))
        self.assertTrue(np.all(
            read_model.bin_edges == self._test_bin_edges
        ))
        self.assertTrue(np.allclose(
                read_model.eval_model_function_density(self._test_x),
                ref_model.eval_model_function_density(self._test_x)
        ))

        if errors:
            _given_error = self._test_parametric_model_with_errors.get_error('test_error')
            _read_error = read_model.get_error('test_error')
            self.assertTrue(_given_error['enabled'] == _read_error['enabled'])
            self.assertTrue(
                np.allclose(
                    _given_error['err'].error,
                    _read_error['err'].error
                )
            )
            self.assertTrue(_given_error['err'].corr_coeff == _read_error['err'].corr_coeff)

    @staticmethod
    def _get_model_from_string(input_string):
        _testfile_stringstream = IOStreamHandle(StringIO(input_string))
        return ParametricModelYamlReader(_testfile_stringstream).read()

    @staticmethod
    def _get_model_from_roundtrip(model):
        _roundtrip_stringstream = IOStreamHandle(StringIO())
        _roundtrip_streamreader = ParametricModelYamlReader(_roundtrip_stringstream)
        _roundtrip_streamwriter = ParametricModelYamlWriter(model, _roundtrip_stringstream)
        _roundtrip_streamwriter.write()
        _roundtrip_stringstream.seek(0)  # return to beginning
        return _roundtrip_streamreader.read()


    @staticmethod
    def linear_model(x, a, b):
        return a * x + b

    @staticmethod
    def linear_model_antiderivative(x, a, b):
        return 0.5 * a * x ** 2 + b * x

    @staticmethod
    @np.vectorize
    def linear_model_numpy_vectorize(x, a, b):
        return 0.5 * a * x ** 2 + b * x

    def setUp(self):
        self._test_n_bins = 5
        self._test_bin_range = (0, 5)
        self._test_bin_edges = np.arange(6)
        self._test_model_parameters = np.array([0.0, 0.08])
        self._test_x = np.linspace(start=0.0, stop=5.0, num=11, endpoint=True)
        self._test_parametric_model = HistParametricModel(
            n_bins=self._test_n_bins, 
            bin_range=self._test_bin_range,
            model_density_func=HistModelFunction(
                TestHistParametricModelYamlRepresenter.linear_model),
            model_parameters=self._test_model_parameters
        )

        self._test_parametric_model_with_errors = HistParametricModel(
            n_bins=self._test_n_bins, 
            bin_range=self._test_bin_range,
            model_density_func=HistModelFunction(
                TestHistParametricModelYamlRepresenter.linear_model),
            model_parameters=self._test_model_parameters
        )
        self._test_parametric_model_with_errors.add_error(
            err_val=0.1, 
            name='test_error', 
            correlation=0, 
            relative=False
        )
        
        self._test_parametric_model_antiderivative = HistParametricModel(
            n_bins=self._test_n_bins,
            bin_range=self._test_bin_range,
            model_density_func=HistModelFunction(
                TestHistParametricModelYamlRepresenter.linear_model),
            model_parameters=self._test_model_parameters,
            bin_evaluation=TestHistParametricModelYamlRepresenter.linear_model_antiderivative
        )

        self._test_parametric_model_numpy_vectorize = HistParametricModel(
            n_bins=self._test_n_bins,
            bin_range=self._test_bin_range,
            model_density_func=HistModelFunction(
                TestHistParametricModelYamlRepresenter.linear_model),
            model_parameters=self._test_model_parameters,
            bin_evaluation=TestHistParametricModelYamlRepresenter.linear_model_numpy_vectorize
        )

    def test_read_from_testfile_stream(self):
        _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(
            TEST_PARAMETRIC_MODEL_HIST)
        self._assert_read_model_correct(_read_parametric_model, self._test_parametric_model)

    def test_read_from_testfile_stream_antiderivative(self):
        _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(
            TEST_PARAMETRIC_MODEL_HIST_ANTIDERIVATIVE)
        self._assert_read_model_correct(
            _read_parametric_model, self._test_parametric_model_antiderivative)

    def test_read_from_testfile_stream_numpy_vectorize(self):
        _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(
            TEST_PARAMETRIC_MODEL_HIST_NUMPY_VECTORIZE)
        self._assert_read_model_correct(
            _read_parametric_model, self._test_parametric_model_numpy_vectorize)

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(
                TEST_PARAMETRIC_MODEL_HIST_MISSING_KEYWORD)

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(
                TEST_PARAMETRIC_MODEL_HIST_EXTRA_KEYWORD)

    def test_read_from_testfile_stream_with_errors(self):
        _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(
            TEST_PARAMETRIC_MODEL_HIST_WITH_ERRORS)
        self._assert_read_model_correct(
            _read_parametric_model, self._test_parametric_model_with_errors, errors=True)

    def test_read_from_testfile_stream_legacy(self):
        _read_parametric_model = TestHistParametricModelYamlRepresenter._get_model_from_string(
            TEST_PARAMETRIC_MODEL_HIST_LEGACY)
        self._assert_read_model_correct(_read_parametric_model, self._test_parametric_model)

    def test_round_trip_with_stringstream(self):
        _read_parametric_model = self._get_model_from_roundtrip(self._test_parametric_model)
        self._assert_read_model_correct(_read_parametric_model, self._test_parametric_model)

    def test_round_trip_with_stringstream_antiderivative(self):
        _read_parametric_model = self._get_model_from_roundtrip(
            self._test_parametric_model_antiderivative)
        self._assert_read_model_correct(
            _read_parametric_model, self._test_parametric_model_antiderivative)

    def test_round_trip_with_stringstream_numpy_vectorize(self):
        _read_parametric_model = self._get_model_from_roundtrip(
            self._test_parametric_model_numpy_vectorize)
        self._assert_read_model_correct(
            _read_parametric_model, self._test_parametric_model_numpy_vectorize)

    def test_round_trip_with_stringstream_with_errors(self):
        _read_parametric_model = self._get_model_from_roundtrip(
            self._test_parametric_model_with_errors)
        self._assert_read_model_correct(
            _read_parametric_model, self._test_parametric_model_with_errors, errors=True)


TEST_PARAMETRIC_MODEL_INDEXED="""
type: indexed
model_function:
    type: indexed
    python_code: |
        def linear_model(a, b):
            return a * np.arange(6) + b
model_parameters: [1.1, -1.5]
"""

TEST_PARAMETRIC_MODEL_INDEXED_MISSING_KEYWORD="""
type: indexed
model_parameters: [1.1, -1.5]
"""

TEST_PARAMETRIC_MODEL_INDEXED_EXTRA_KEYWORD = TEST_PARAMETRIC_MODEL_INDEXED + """
extra_keyword: 3.14
"""

TEST_PARAMETRIC_MODEL_INDEXED_WITH_ERRORS = TEST_PARAMETRIC_MODEL_INDEXED + """
errors:
  - correlation_coefficient: 0.0
    error_value: 0.1
    name: test_error
    relative: false
    type: simple
"""

class TestIndexedParametricModelYamlRepresenter(unittest.TestCase):

    TEST_X = np.arange(6)
        
    @staticmethod
    def linear_model(a, b):
        return a * np.arange(6) + b
   
    def setUp(self):
        self._test_model_parameters = np.array([1.1, -1.5])
        self._test_parametric_model = IndexedParametricModel(
            IndexedModelFunction(TestIndexedParametricModelYamlRepresenter.linear_model),
            self._test_model_parameters
        )

        self._test_parametric_model_with_errors = IndexedParametricModel(
            IndexedModelFunction(TestIndexedParametricModelYamlRepresenter.linear_model),
            self._test_model_parameters
        )
        self._test_parametric_model_with_errors.add_error(
            err_val=0.1, 
            name='test_x_error', 
            correlation=0, 
            relative=False
        )
        
        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._roundtrip_stringstream_with_errors = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_INDEXED))
        self._testfile_stringstream_with_errors = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_INDEXED_WITH_ERRORS))
        
        self._roundtrip_streamreader = ParametricModelYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamreader_with_errors = ParametricModelYamlReader(self._roundtrip_stringstream_with_errors)
        self._roundtrip_streamwriter = ParametricModelYamlWriter(
            self._test_parametric_model, 
            self._roundtrip_stringstream
        )
        self._roundtrip_streamwriter_with_errors = ParametricModelYamlWriter(
            self._test_parametric_model_with_errors, 
            self._roundtrip_stringstream_with_errors
        )
        self._testfile_streamreader = ParametricModelYamlReader(self._testfile_stringstream)
        self._testfile_streamreader_with_errors = ParametricModelYamlReader(self._testfile_stringstream_with_errors)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_INDEXED_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_INDEXED_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ParametricModelYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ParametricModelYamlReader(self._testfile_stringstream_extra_keyword)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_write_to_roundtrip_stringstream_with_errors(self):
        self._roundtrip_streamwriter_with_errors.write()

    def test_read_from_testfile_stream(self):
        _read_parametric_model = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_parametric_model, IndexedParametricModel))
        
        self.assertTrue(
            np.allclose(
                _read_parametric_model.parameters,
                self._test_model_parameters
            )
        )
        self.assertTrue(
            np.allclose(
                _read_parametric_model.eval_model_function(),
                self._test_parametric_model.eval_model_function()
            )
        )
        
    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_read_from_testfile_stream_with_errors(self):
        _read_parametric_model = self._testfile_streamreader_with_errors.read()
        self.assertTrue(isinstance(_read_parametric_model, IndexedParametricModel))
        
        self.assertTrue(
            np.allclose(
                _read_parametric_model.parameters,
                self._test_model_parameters
            )
        )
        self.assertTrue(
            np.allclose(
                _read_parametric_model.eval_model_function(),
                self._test_parametric_model_with_errors.eval_model_function()
            )
        )
        
        _given_error = self._test_parametric_model_with_errors.get_error('test_x_error')
        _read_error = _read_parametric_model.get_error('test_error')
        self.assertTrue(_given_error['enabled'] == _read_error['enabled'])
        self.assertTrue(
            np.allclose(
                _given_error['err'].error,
                _read_error['err'].error
            )
        )
        self.assertTrue(_given_error['err'].corr_coeff == _read_error['err'].corr_coeff)
        
    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_parametric_model = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_parametric_model, IndexedParametricModel))
        
        self.assertTrue(
            np.allclose(
                _read_parametric_model.parameters,
                self._test_model_parameters
            )
        )
        self.assertTrue(
            np.allclose(
                _read_parametric_model.eval_model_function(),
                self._test_parametric_model.eval_model_function()
            )
        )
        
    def test_round_trip_with_stringstream_with_errors(self):
        self._roundtrip_streamwriter_with_errors.write()
        self._roundtrip_stringstream_with_errors.seek(0)  # return to beginning
        _read_parametric_model = self._roundtrip_streamreader_with_errors.read()
        self.assertTrue(isinstance(_read_parametric_model, IndexedParametricModel))
        
        self.assertTrue(
            np.allclose(
                _read_parametric_model.parameters,
                self._test_model_parameters
            )
        )
        self.assertTrue(
            np.allclose(
                _read_parametric_model.eval_model_function(),
                self._test_parametric_model_with_errors.eval_model_function()
            )
        )
        
        _given_error = self._test_parametric_model_with_errors.get_error('test_x_error')
        _read_error = _read_parametric_model.get_error('test_x_error')
        self.assertTrue(_given_error['enabled'] == _read_error['enabled'])
        self.assertTrue(
            np.allclose(
                _given_error['err'].error,
                _read_error['err'].error
            )
        )
        self.assertTrue(_given_error['err'].corr_coeff == _read_error['err'].corr_coeff)

TEST_PARAMETRIC_MODEL_XY="""
type: xy
x_data: [0.0, 1.0, 2.0, 3.0, 4.0, 5.0]
model_function:
    type: xy
    python_code: |
        def linear_model(x, a, b):
            return a * x + b
model_parameters: [1.1, -1.5]
"""

TEST_PARAMETRIC_MODEL_XY_MISSING_KEYWORD="""
type: xy
model_function:
    type: xy
    python_code: |
        def linear_model(x, a, b):
            return a * x + b
model_parameters: [1.1, -1.5]
"""

TEST_PARAMETRIC_MODEL_XY_EXTRA_KEYWORD = TEST_PARAMETRIC_MODEL_XY + """
extra_keyword: 3.14
"""

TEST_PARAMETRIC_MODEL_XY_WITH_ERRORS = TEST_PARAMETRIC_MODEL_XY + """
x_errors:
  - correlation_coefficient: 0.0
    error_value: 0.1
    name: test_x_error
    relative: false
    type: simple
"""

class TestXYParametricModelYamlRepresenter(unittest.TestCase):

    @staticmethod
    def linear_model(x, a, b):
        return a * x + b
   
    def setUp(self):
        self._test_x = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        self._test_model_parameters = np.array([1.1, -1.5])
        self._test_parametric_model = XYParametricModel(
            self._test_x, 
            ModelFunctionBase(TestXYParametricModelYamlRepresenter.linear_model),
            self._test_model_parameters
        )

        self._test_parametric_model_with_errors = XYParametricModel(
            self._test_x, 
            ModelFunctionBase(TestXYParametricModelYamlRepresenter.linear_model),
            self._test_model_parameters
        )
        self._test_parametric_model_with_errors.add_error(
            axis='x',
            err_val=0.1, 
            name='test_x_error', 
            correlation=0, 
            relative=False
        )
        
        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._roundtrip_stringstream_with_errors = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_XY))
        self._testfile_stringstream_with_errors = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_XY_WITH_ERRORS))
        
        self._roundtrip_streamreader = ParametricModelYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamreader_with_errors = ParametricModelYamlReader(self._roundtrip_stringstream_with_errors)
        self._roundtrip_streamwriter = ParametricModelYamlWriter(
            self._test_parametric_model, 
            self._roundtrip_stringstream
        )
        self._roundtrip_streamwriter_with_errors = ParametricModelYamlWriter(
            self._test_parametric_model_with_errors, 
            self._roundtrip_stringstream_with_errors
        )
        self._testfile_streamreader = ParametricModelYamlReader(self._testfile_stringstream)
        self._testfile_streamreader_with_errors = ParametricModelYamlReader(self._testfile_stringstream_with_errors)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_XY_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_PARAMETRIC_MODEL_XY_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ParametricModelYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ParametricModelYamlReader(self._testfile_stringstream_extra_keyword)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_write_to_roundtrip_stringstream_with_errors(self):
        self._roundtrip_streamwriter_with_errors.write()

    def test_read_from_testfile_stream(self):
        _read_parametric_model = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_parametric_model, XYParametricModel))
        
        self.assertTrue(
            np.allclose(
                _read_parametric_model.parameters,
                self._test_model_parameters
            )
        )
        _read_parametric_model.x = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        self.assertTrue(
            np.allclose(
                _read_parametric_model.y,
                self._test_parametric_model.y
            )
        )
        
    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_read_from_testfile_stream_with_errors(self):
        _read_parametric_model = self._testfile_streamreader_with_errors.read()
        self.assertTrue(isinstance(_read_parametric_model, XYParametricModel))
        
        self.assertTrue(
            np.allclose(
                _read_parametric_model.parameters,
                self._test_model_parameters
            )
        )
        _read_parametric_model.x = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        self.assertTrue(
            np.allclose(
                _read_parametric_model.y,
                self._test_parametric_model_with_errors.y
            )
        )
        
        _given_error = self._test_parametric_model_with_errors.get_error('test_x_error')
        _read_error = _read_parametric_model.get_error('test_x_error')
        self.assertTrue(_given_error['axis'] == _read_error['axis'])
        self.assertTrue(_given_error['enabled'] == _read_error['enabled'])
        self.assertTrue(
            np.allclose(
                _given_error['err'].error,
                _read_error['err'].error
            )
        )
        self.assertTrue(_given_error['err'].corr_coeff == _read_error['err'].corr_coeff)
        
    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_parametric_model = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_parametric_model, XYParametricModel))
        
        self.assertTrue(
            np.allclose(
                _read_parametric_model.parameters,
                self._test_model_parameters
            )
        )
        _read_parametric_model.x = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        self.assertTrue(
            np.allclose(
                _read_parametric_model.y,
                self._test_parametric_model.y
            )
        )
        
    def test_round_trip_with_stringstream_with_errors(self):
        self._roundtrip_streamwriter_with_errors.write()
        self._roundtrip_stringstream_with_errors.seek(0)  # return to beginning
        _read_parametric_model = self._roundtrip_streamreader_with_errors.read()
        self.assertTrue(isinstance(_read_parametric_model, XYParametricModel))
        
        self.assertTrue(
            np.allclose(
                _read_parametric_model.parameters,
                self._test_model_parameters
            )
        )
        _read_parametric_model.x = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        self.assertTrue(
            np.allclose(
                _read_parametric_model.y,
                self._test_parametric_model_with_errors.y
            )
        )
        
        _given_error = self._test_parametric_model_with_errors.get_error('test_x_error')
        _read_error = _read_parametric_model.get_error('test_x_error')
        self.assertTrue(_given_error['axis'] == _read_error['axis'])
        self.assertTrue(_given_error['enabled'] == _read_error['enabled'])
        self.assertTrue(
            np.allclose(
                _given_error['err'].error,
                _read_error['err'].error
            )
        )
        self.assertTrue(_given_error['err'].corr_coeff == _read_error['err'].corr_coeff)



from .._yaml_base import YamlWriterMixin, YamlReaderMixin
from ._base import ConstraintDReprBase
from .. import _AVAILABLE_REPRESENTATIONS
from ....core.constraint import GaussianSimpleParameterConstraint, GaussianMatrixParameterConstraint

__all__ = ['ConstraintYamlWriter', 'ConstraintYamlReader']


class ConstraintYamlWriter(YamlWriterMixin, ConstraintDReprBase):

    def __init__(self, constraint, output_io_handle):
        super(ConstraintYamlWriter, self).__init__(
            output_io_handle=output_io_handle,
            constraint=constraint)

    @classmethod
    def _make_representation(cls, constraint):
        _yaml_doc = dict()

        # -- determine constraint type
        _class = constraint.__class__
        _type = cls._CLASS_TO_OBJECT_TYPE_NAME.get(_class, None)
        if _type is None:
            raise TypeError("Constraint type unknown or not supported: %s" % constraint.__class__)
        _yaml_doc['type'] = _type

        if _class is GaussianSimpleParameterConstraint:
            _yaml_doc['index'] = constraint.index
            _yaml_doc['value'] = constraint.value
            _yaml_doc['uncertainty'] = constraint.uncertainty
            _yaml_doc['relative'] = constraint.relative
        elif _class is GaussianMatrixParameterConstraint:
            _yaml_doc['indices'] = constraint.indices.tolist()
            _yaml_doc['values'] = constraint.values.tolist()
            if constraint.relative:
                if constraint.matrix_type == 'cov':
                    _yaml_doc['matrix'] = constraint.cov_mat_rel.tolist()
                else:
                    _yaml_doc['matrix'] = constraint.cor_mat.tolist()
                    _yaml_doc['uncertainties'] = constraint.uncertainties_rel.tolist()
            else:
                if constraint.matrix_type == 'cov':
                    _yaml_doc['matrix'] = constraint.cov_mat.tolist()
                else:
                    _yaml_doc['matrix'] = constraint.cor_mat.tolist()
                    _yaml_doc['uncertainties'] = constraint.uncertainties.tolist()
            _yaml_doc['relative'] = constraint.relative
            _yaml_doc['matrix_type'] = constraint.matrix_type
        else:
            raise TypeError('Unknown constraint class: %s' % _class)

        return _yaml_doc


class ConstraintYamlReader(YamlReaderMixin, ConstraintDReprBase):

    def __init__(self, input_io_handle):
        super(ConstraintYamlReader, self).__init__(
            input_io_handle=input_io_handle,
            constraint=None)

    @classmethod
    def _get_required_keywords(cls, yaml_doc, constraint_class):
        if constraint_class is GaussianSimpleParameterConstraint:
            return ['index', 'value', 'uncertainty']
        if constraint_class is GaussianMatrixParameterConstraint:
            return ['indices', 'values', 'matrix']
        raise TypeError('Unknown constraint class: %s' % constraint_class)

    @classmethod
    def _modify_yaml_doc(cls, yaml_doc, kafe_object_class, parameter_names=None, **kwargs):
        if kafe_object_class is GaussianSimpleParameterConstraint:
            _par_name = yaml_doc.pop('name', None)
            if _par_name:
                yaml_doc['index'] = parameter_names.index(_par_name)
        elif kafe_object_class is GaussianMatrixParameterConstraint:
            _par_names = yaml_doc.pop('names', None)
            if _par_names:
                yaml_doc['indices'] = [parameter_names.index(_par_name) for _par_name in _par_names]
        return yaml_doc

    @classmethod
    def _convert_yaml_doc_to_object(cls, yaml_doc):
        # -- determine constraint class from type
        _constraint_type = yaml_doc.pop('type', 'simple')
        _class = cls._OBJECT_TYPE_NAME_TO_CLASS.get(_constraint_type, None)

        if _class is GaussianSimpleParameterConstraint:
            _constraint_object = GaussianSimpleParameterConstraint(
                index=yaml_doc.pop('index'),
                value=yaml_doc.pop('value'),
                uncertainty=yaml_doc.pop('uncertainty'),
                relative=yaml_doc.pop('relative', False)
            )
        elif _class is GaussianMatrixParameterConstraint:
            _constraint_object = GaussianMatrixParameterConstraint(
                indices=yaml_doc.pop('indices'),
                values=yaml_doc.pop('values'),
                matrix=yaml_doc.pop('matrix'),
                matrix_type=yaml_doc.pop('matrix_type', 'cov'),
                uncertainties=yaml_doc.pop('uncertainties', None),
                relative=yaml_doc.pop('relative', False)
            )
        else:
            raise TypeError('Unknown constraint class: %s' % _class)
        return _constraint_object, yaml_doc


# register the above classes in the module-level dictionary
ConstraintYamlReader._register_class(_AVAILABLE_REPRESENTATIONS)
ConstraintYamlWriter._register_class(_AVAILABLE_REPRESENTATIONS)



import numpy as np
import six
import unittest

from kafe2.core.minimizers import AVAILABLE_MINIMIZERS
from kafe2.core.fitters.nexus import Nexus, Parameter, Function
from kafe2.core.fitters.nexus_fitter import NexusFitter


class AbstractTestNexusFitter(object):

    MINIMIZER = None

    @staticmethod
    def slsq(x, y, x_1, y_1, x_2, y_2):
        return (
            (x - x_1)**2 + (y - y_1)**2 +
            (x - x_2)**2 + (y - y_2)**2
        )

    def setUp(self):
        self.nexus = Nexus()

        self._ref_xy_1 = (3, 9)
        self._ref_xy_2 = (7, 11)
        self._ref_xy = tuple(
            0.5*(np.array(self._ref_xy_1) + np.array(self._ref_xy_2))
        )
        self._startval_x = 300
        self._startval_y = -20

        par_x_1, par_y_1 = (
            self.nexus.add(
                Parameter(self._ref_xy_1[0], name='x_1')
            ),
            self.nexus.add(
                Parameter(self._ref_xy_1[1], name='y_1')
            )
        )
        par_x_2, par_y_2 = (
            self.nexus.add(
                Parameter(self._ref_xy_2[0], name='x_2')
            ),
            self.nexus.add(
                Parameter(self._ref_xy_2[1], name='y_2')
            )
        )

        par_x, par_y = (
            self.nexus.add(
                Parameter(self._startval_x, name='x')
            ),
            self.nexus.add(
                Parameter(self._startval_y, name='y')
            )
        )

        self.nexus.add_function(self.slsq)

        self.fitter = NexusFitter(
            self.nexus,
            parameters_to_fit=('x', 'y'),
            parameter_to_minimize='slsq',
            minimizer=self.MINIMIZER
        )

    def _assert_fit_results(self, xy_val=None):

        xy_val = xy_val or self._ref_xy

        for _par_name, _val in zip(('x', 'y'), xy_val):
            self.assertAlmostEqual(
                self.fitter.get_fit_parameter_values()[_par_name],
                _val,
                places=3
            )

        _slsq_args = tuple(xy_val) + self._ref_xy_1 + self._ref_xy_2
        self.assertAlmostEqual(
            self.fitter.parameter_to_minimize_value,
            self.slsq(*_slsq_args),
            places=2
        )

    # -- init

    def test_init_inexistent_fit_parameters_raise(self):
        with self.assertRaises(ValueError):
            NexusFitter(
                self.nexus,
                parameters_to_fit=('bogus_parameter',),
                parameter_to_minimize='slsq',
            )

    def test_init_inexistent_min_parameter_raise(self):
        with self.assertRaises(ValueError):
            NexusFitter(
                self.nexus,
                parameters_to_fit=('x', 'y'),
                parameter_to_minimize='bogus_parameter',
            )

    # -- basic interface

    def test_simple_properties(self):
        self.assertEqual(
            self.fitter.parameters_to_fit,
            ('x', 'y')
        )
        self.assertEqual(
            self.fitter.parameter_to_minimize,
            'slsq'
        )
        self.assertEqual(
            self.fitter.get_fit_parameter_values(),
            {
                'x': self._startval_x,
                'y': self._startval_y,
            }
        )
        self.assertEqual(
            self.fitter.n_fit_par,
            2
        )
        self._assert_fit_results(xy_val=[self._startval_x, self._startval_y])

    def test_set_all_fit_parameter_values(self):
        self.fitter.set_all_fit_parameter_values([7, 2])
        self._assert_fit_results(xy_val=[7, 2])

    def test_set_all_fit_parameter_values_length_mismatch_raise(self):
        with self.assertRaises(ValueError):
            self.fitter.set_all_fit_parameter_values([7, 2, 30])
        with self.assertRaises(ValueError):
            self.fitter.set_all_fit_parameter_values([2])

    def test_set_fit_parameter_values(self):
        self.fitter.set_fit_parameter_values(x=7, y=2)
        self._assert_fit_results(xy_val=[7, 2])

    def test_set_fit_parameter_values_unknown_raise(self):
        with self.assertRaises(ValueError):
            self.fitter.set_fit_parameter_values(bogus_parameter=7)

    def test_do_fit(self):
        self.fitter.do_fit()
        self._assert_fit_results()  # nominal fit results

    def test_state_is_from_minimizer(self):
        self.assertEqual(self.fitter.state_is_from_minimizer, False)
        self.fitter.do_fit()
        self.assertEqual(self.fitter.state_is_from_minimizer, True)
        self.fitter.set_fit_parameter_values(x=7)
        self.assertEqual(self.fitter.state_is_from_minimizer, False)

    # -- parameter errors

    def test_fit_parameter_errors(self):
        self.fitter.do_fit()
        self.assertAlmostEqual(
            self.fitter.fit_parameter_errors[0],
            1.0/np.sqrt(2)
        )
        self.assertAlmostEqual(
            self.fitter.fit_parameter_errors[1],
            1.0/np.sqrt(2)
        )

    def test_fit_parameter_cov_mat(self):
        self.fitter.do_fit()
        self.assertAlmostEqual(
            self.fitter.fit_parameter_cov_mat[0, 0],
            0.5
        )

    def test_fit_parameter_cor_mat(self):
        self.fitter.do_fit()
        self.assertAlmostEqual(
            self.fitter.fit_parameter_cor_mat[0, 0],
            1.0
        )

    def test_asymmetric_fit_parameter_errors(self):
        self.fitter.do_fit()
        self.assertAlmostEqual(
            self.fitter.asymmetric_fit_parameter_errors[0][0],
            -1.0/np.sqrt(2),
            places=4
        )
        self.assertAlmostEqual(
            self.fitter.asymmetric_fit_parameter_errors[0][1],
            1.0/np.sqrt(2),
            places=4
        )

    def test_asymmetric_fit_parameter_errors_if_calculated(self):
        self.fitter.do_fit()
        self.assertIs(
            self.fitter.asymmetric_fit_parameter_errors_if_calculated,
            None
        )

    # -- fixing & limiting parameters

    def test_fix_parameter(self):
        self.fitter.fix_parameter('x')
        self.fitter.do_fit()
        self._assert_fit_results(xy_val=[self._startval_x, self._ref_xy[1]])

    def test_fix_parameter_to_value(self):
        self.fitter.fix_parameter('x', 200)
        self.fitter.do_fit()
        self._assert_fit_results(xy_val=[200, self._ref_xy[1]])

    def test_release_parameter(self):
        self.fitter.fix_parameter('x')
        self.fitter.do_fit()
        self.fitter.release_parameter('x')
        self.fitter.do_fit()
        self._assert_fit_results()  # nominal fit results

    def test_limit_parameter(self):
        self.fitter.limit_parameter('x', (6, 10))
        self.fitter.do_fit()
        self._assert_fit_results(xy_val=[6, self._ref_xy[1]])

    def test_unlimit_parameter(self):
        self.fitter.limit_parameter('x', (6, 10))
        self.fitter.do_fit()
        self.fitter.unlimit_parameter('x')
        self.fitter.do_fit()
        self._assert_fit_results()  # nominal fit results


    # -- profiles & contours

    def test_profile_without_do_fit(self):
        with self.assertRaises(RuntimeError):
            self.fitter.profile('x')

    def test_profile(self):
        self.fitter.do_fit()
        p = self.fitter.profile('x', bins=20, bound=2)
        self.assertEqual(
            p.shape,
            (2, 20)
        )
        self.assertAlmostEqual(
            p[0][-1] - p[0][0],
            4 * self.fitter.fit_parameter_errors[0]
        )

    def test_contour_without_do_fit(self):
        with self.assertRaises(RuntimeError):
            self.fitter.contour('x', 'y')

    def test_contour(self):
        self.fitter.do_fit()
        c = self.fitter.contour('x', 'y')
        # TODO: test result?


if 'scipy' in AVAILABLE_MINIMIZERS:
    class TestNexusFitterScipy(AbstractTestNexusFitter, unittest.TestCase):
        MINIMIZER = 'scipy'

if 'iminuit' in AVAILABLE_MINIMIZERS:
    class TestNexusFitterIMinuit(AbstractTestNexusFitter, unittest.TestCase):
        MINIMIZER = 'iminuit'



import logging
import numpy as np
from copy import deepcopy

from .minimizer_base import MinimizerBase
from ..contour import ContourFactory

try:
    import iminuit
except (ImportError, SyntaxError):
    raise


_IMINUIT_1 = int(iminuit.__version__[0]) < 2


class MinimizerIMinuit(MinimizerBase):
    def __init__(self,
                 parameter_names, parameter_values, parameter_errors,
                 function_to_minimize, tolerance=1e-2, errordef=iminuit.Minuit.LEAST_SQUARES,
                 strategy=1):
        self._strategy = strategy

        # initialize the minimizer parameter specification
        self._minimizer_param_dict = {}
        for _pn in parameter_names:
            self._minimizer_param_dict["fix_" + _pn] = False
            self._minimizer_param_dict["limit_" + _pn] = None

        self.reset()  # sets self.__iminuit and caches to None
        super(MinimizerIMinuit, self).__init__(
            parameter_names=parameter_names, parameter_values=parameter_values,
            parameter_errors=parameter_errors, function_to_minimize=function_to_minimize,
            tolerance=tolerance, errordef=errordef
        )

    # -- private methods

    def _invalidate_cache(self):
        self._par_val = None
        self._par_err = None
        self._fmin_struct = None
        super(MinimizerIMinuit, self)._invalidate_cache()

    def _save_state(self):
        if self._par_val is None:
            self._save_state_dict["par_val"] = self._par_val
        else:
            self._save_state_dict["par_val"] = np.array(self._par_val)
        if self._par_err is None:
            self._save_state_dict["par_err"] = self._par_err
        else:
            self._save_state_dict["par_err"] = np.array(self._par_err)
        self._save_state_dict["fmin_struct"] = deepcopy(self._fmin_struct)
        self._save_state_dict['minimizer_param_dict'] = self._minimizer_param_dict
        self._save_state_dict['iminuit'] = self.__iminuit
        super(MinimizerIMinuit, self)._save_state()

    def _load_state(self):
        self.reset()
        self._par_val = self._save_state_dict["par_val"]
        if self._par_val is not None:
            self._par_val = np.array(self._par_val)
        self._par_err = self._save_state_dict["par_err"]
        if self._par_err is not None:
            self._par_err = np.array(self._par_err)
        self._fmin_struct = deepcopy(self._save_state_dict["fmin_struct"])
        self._minimizer_param_dict = self._save_state_dict['minimizer_param_dict']
        self.__iminuit = self._save_state_dict['iminuit']
        self._func_handle(*self.parameter_values)  # call the function to propagate the changes to the nexus
        super(MinimizerIMinuit, self)._load_state()

    def _get_fmin_struct(self):
        if self._fmin_struct is None:
            self._fmin_struct = self._get_iminuit().get_fmin()
        return self._fmin_struct

    def _get_iminuit(self):
        if self.__iminuit is None:
            if _IMINUIT_1:
                self.__iminuit = iminuit.Minuit(self._func_wrapper,
                                                forced_parameters=self.parameter_names,
                                                errordef=self.errordef,
                                                **self._minimizer_param_dict)
            else:
                _parameter_values = [
                    self._minimizer_param_dict[_pn] for _pn in self.parameter_names]
                self.__iminuit = iminuit.Minuit(
                    self._func_wrapper, *_parameter_values,
                    name=self.parameter_names
                )
                for _i, _par_name_i in enumerate(self.parameter_names):
                    self.__iminuit.fixed[_i] = self._minimizer_param_dict["fix_" + _par_name_i]
                    self.__iminuit.limits[_i] = self._minimizer_param_dict["limit_" + _par_name_i]
                    _par_err_i = self._minimizer_param_dict.get("error_" + _par_name_i, None)
                    if _par_err_i is not None:
                        self.__iminuit.errors[_i] = _par_err_i
                self.__iminuit.errordef = self.errordef
            # set logging level in iminuit arcording to the root logger
            if logging.root.level <= logging.DEBUG:
                self.__iminuit.print_level = 2
            elif logging.root.level <= logging.INFO:
                self.__iminuit.print_level = 1
            else:  # default is logging.WARN, show nothing
                self.__iminuit.print_level = 0
            self.__iminuit.strategy = self._strategy
            self.__iminuit.tol = self.tolerance
        return self.__iminuit

    def _calculate_asymmetric_parameter_errors(self):
        try:
            if _IMINUIT_1:
                _minos_result = self._get_iminuit().minos()
            else:
                _m = self._get_iminuit()
                _m.minos()
                _minos_result = _m.merrors
                _par_names_free = [_pn for _pn in self.parameter_names if not self.is_fixed(_pn)]
        except RuntimeError:
            return None
        _asymm_par_errs = np.zeros(shape=(self.num_pars, 2))
        for _par_name in self.parameter_names:
            _par_index = self.parameter_names.index(_par_name)
            if self.is_fixed(_par_name):
                _asymm_par_errs[_par_index, :] = 0.0
            else:
                if _IMINUIT_1:
                    _asymm_par_errs[_par_index, 0] = _minos_result[_par_name]['lower']
                    _asymm_par_errs[_par_index, 1] = _minos_result[_par_name]['upper']
                else:
                    _par_index_free = _par_names_free.index(_par_name)
                    _asymm_par_errs[_par_index, 0] = _minos_result[_par_index_free].lower
                    _asymm_par_errs[_par_index, 1] = _minos_result[_par_index_free].upper
        self.minimize()
        return _asymm_par_errs

    # -- public properties

    @property
    def hessian(self):
        if not self.did_fit:
            return None
        if self._hessian is None:
            _cov_mat = self.cov_mat
            if _cov_mat is None:
                self._hessian = None
            else:
                _submat = self._remove_zeroes_for_fixed(_cov_mat)
                _submat_inv = 2.0 * self.errordef * np.linalg.inv(_submat)
                self._hessian = self._fill_in_zeroes_for_fixed(_submat_inv)
        return None if self._hessian is None else self._hessian.copy()

    @property
    def cov_mat(self):
        if not self.did_fit:
            return None
        if self._par_cov_mat is None:
            self._save_state()
            try:
                self._get_iminuit().hesse()
                if _IMINUIT_1:
                    # FIX_UPSTREAM we need skip_fixed=False, but this is unsupported
                    # _mat = self._get_iminuit().matrix(correlation, skip_fixed=False)

                    # ... so use skip_fixed=True instead and fill in the gaps
                    _mat = self._get_iminuit().matrix(correlation=False, skip_fixed=True)
                    _mat = np.asarray(_mat)  # reshape into numpy matrix
                    _mat = self._fill_in_zeroes_for_fixed(_mat)  # fill in fixed par 'gaps'
                else:
                    # iminuit 2 already fills in zeros for fixed parameters.
                    _mat = self._get_iminuit().covariance
                    if _mat is not None:
                        _mat = np.asarray(_mat)
                self._func_wrapper_unpack_args(self.parameter_values)
            except RuntimeError:
                _mat = None
            self._load_state()
            self._par_cov_mat = _mat
        return None if self._par_cov_mat is None else self._par_cov_mat.copy()

    @property
    def hessian_inv(self):
        if not self.did_fit:
            return None
        if self._hessian_inv is None:
            self._hessian_inv = self.cov_mat / (2.0 * self.errordef)
        return None if self._hessian_inv is None else self._hessian_inv.copy()

    @property
    def parameter_values(self):
        if self._par_val is None:
            _m = self._get_iminuit()
            if _IMINUIT_1:
                if not _m.is_clean_state():
                    # if the fit has been performed at least once
                    _param_struct = _m.get_param_states()
                    _pvals = [p.value for p in _param_struct]
                else:
                    # need to hack to get initial parameter values
                    _v = _m.values
                    _pvals = [_v[pname] for pname in self.parameter_names]
                self._par_val = np.array(_pvals)
            else:
                self._par_val = np.array(_m.values)
        return self._par_val.copy()

    @parameter_values.setter
    def parameter_values(self, new_values):
        for _pn, _pv, in zip(self._par_names, new_values):
            self._minimizer_param_dict[_pn] = _pv
        self.reset()

    @property
    def parameter_errors(self):
        if self._par_err is None:
            _m = self._get_iminuit()
            if _IMINUIT_1:
                if not _m.is_clean_state():
                    # if the fit has been performed at least once
                    _param_struct = _m.get_param_states()
                    self._par_err = np.array(
                        [_par.error if not self.is_fixed(_par_name) else 0.0
                         for _par, _par_name in zip(_param_struct, self.parameter_names)])
                else:
                    # need to hack to get initial parameter errors
                    _e = _m.errors
                    self._par_err = np.array(
                        [_e[_par_name] if not self.is_fixed(_par_name) else 0.0
                         for _par_name in _e])
            else:
                self._par_err = np.array(_m.errors)
                # Explicitly set errors of fixed parameters to 0:
                for _i, _par_name_i in enumerate(self.parameter_names):
                    if self.is_fixed(_par_name_i):
                        self._par_err[_i] = 0.0
        return self._par_err.copy()

    @parameter_errors.setter
    def parameter_errors(self, new_errors):
        if not np.all(np.array(new_errors) > 0):
            raise ValueError("All parameter errors must be > 0! Received: %s" % new_errors)
        for _pn, _pe in zip(self._par_names, new_errors):
            self._minimizer_param_dict["error_" + _pn] = _pe
        self.reset()

    # -- private "properties"

    @property
    def _fmin_edm(self):
        _fmin = self._get_fmin_struct()
        return _fmin.edm

    @property
    def _fmin_up(self):
        _fmin = self._get_fmin_struct()
        return _fmin.up

    @property
    def _fmin_has_covariance(self):
        _fmin = self._get_fmin_struct()
        return _fmin.has_covariance

    @property
    def _fmin_covariance_is_pos_def(self):
        _fmin = self._get_fmin_struct()
        return _fmin.has_made_posdef_covar

    @property
    def _fmin_covariance_is_accurate(self):
        _fmin = self._get_fmin_struct()
        return _fmin.has_accurate_covar

    # -- public methods

    def reset(self):
        super(MinimizerIMinuit, self).reset()
        self.__iminuit = None

    def contour(self, parameter_name_1, parameter_name_2, sigma=1.0, **minimizer_contour_kwargs):
        if not self.did_fit:
            raise RuntimeError("Need to perform a fit before calling contour()!")
        _numpoints = minimizer_contour_kwargs.pop("numpoints", 100)
        if minimizer_contour_kwargs:
            raise ValueError(
                "Unknown keyword arguments for contour(): {}".format(minimizer_contour_kwargs.keys()))
        if _IMINUIT_1:
            _x_errs, _y_errs, _contour_line = self._get_iminuit().mncontour(
                parameter_name_1, parameter_name_2, numpoints=_numpoints, sigma=sigma)
        else:
            # The following conversion is derived by integrating the two-dimensional standard
            # normal distribution over a circle of radius sigma centered on (0, 0).
            _cl = 1.0 - np.exp(-0.5 * sigma ** 2)
            _contour_line = self._get_iminuit().mncontour(
                parameter_name_1, parameter_name_2, size=_numpoints, cl=_cl)
        self.minimize()  # return to minimum
        if len(_contour_line) == 0:
            return None  # failed to find any point on contour
        return ContourFactory.create_xy_contour(np.array(_contour_line), sigma)

    def profile(self, parameter_name, bins=20, bound=2, subtract_min=False):
        if not self.did_fit:
            raise RuntimeError("Need to perform a fit before calling profile()!")
        _kwargs = dict(bound=bound, subtract_min=subtract_min)
        if _IMINUIT_1:
            _kwargs["bins"] = bins
        else:
            _kwargs["size"] = bins
        _bins, _vals, _statuses = self.__iminuit.mnprofile(parameter_name, **_kwargs)
        # TODO: check statuses (?)
        self.minimize()  # return to minimum
        return np.array([_bins, _vals])

    def set(self, parameter_name, parameter_value):
        if parameter_name not in self._minimizer_param_dict:
            raise ValueError("No parameter named '%s'!" % (parameter_name,))
        self._minimizer_param_dict[parameter_name] = parameter_value
        self.reset()

    def fix(self, parameter_name):
        self._minimizer_param_dict["fix_" + parameter_name] = True
        if _IMINUIT_1:
            self._get_iminuit().fixed[parameter_name] = True
        else:
            self._get_iminuit().fixed[self.parameter_names.index(parameter_name)] = True
        self._invalidate_cache()

    def is_fixed(self, parameter_name):
        if _IMINUIT_1:
            return self._get_iminuit().fixed[parameter_name]
        else:
            return self._get_iminuit().fixed[self.parameter_names.index(parameter_name)]

    def release(self, parameter_name):
        self._minimizer_param_dict["fix_" + parameter_name] = False
        if _IMINUIT_1:
            self._get_iminuit().fixed[parameter_name] = False
        else:
            self._get_iminuit().fixed[self.parameter_names.index(parameter_name)] = False
        self._invalidate_cache()

    def limit(self, parameter_name, parameter_bounds):
        assert len(parameter_bounds) == 2
        self._minimizer_param_dict["limit_" + parameter_name] = (parameter_bounds[0], parameter_bounds[1])
        self.reset()

    def unlimit(self, parameter_name):
        self._minimizer_param_dict["limit_" + parameter_name] = None
        self.reset()

    def minimize(self, max_calls=6000):
        if np.all([self.is_fixed(_par_name) for _par_name in self.parameter_names]):
            raise RuntimeError("Cannot perform a fit if all parameters are fixed!")

        self._get_iminuit().migrad(ncall=max_calls)

        for (_pn, _pv, _pe) in zip(
                self.parameter_names, self.parameter_values, self.parameter_errors):
            self._minimizer_param_dict[_pn] = _pv
            self._minimizer_param_dict["error_" + _pn] = _pe

        # invalidate cache
        self._did_fit = True
        self._invalidate_cache()



import unittest
from kafe2.test.core.minimizers._base import AbstractMinimizerTest
from kafe2.core.minimizers.scipy_optimize_minimizer import MinimizerScipyOptimize


class TestMinimizerScipyOptimize(AbstractMinimizerTest, unittest.TestCase):

    def _get_minimizer(self, parameter_names, parameter_values, parameter_errors,
                       function_to_minimize):
        return MinimizerScipyOptimize(
            parameter_names=parameter_names, parameter_values=parameter_values,
            parameter_errors=parameter_errors, function_to_minimize=function_to_minimize
        )

    @property
    def _expected_tolerance(self):
        return 1e-6



import numpy as np
from matplotlib.collections import LineCollection

from .._base import PlotAdapterBase
from .._aux import add_pad_to_range


__all__ = ["UnbinnedPlotAdapter"]


class UnbinnedPlotAdapter(PlotAdapterBase):

    PLOT_STYLE_CONFIG_DATA_TYPE = 'unbinned'

    PLOT_SUBPLOT_TYPES = dict(
        PlotAdapterBase.PLOT_SUBPLOT_TYPES,
        model_line=dict(
            plot_adapter_method='plot_model_line',
            target_axes='main',
        )
    )
    PLOT_SUBPLOT_TYPES['model']['hide'] = True  # don't show "model" points

    AVAILABLE_X_SCALES = ('linear', 'log')

    def __init__(self, unbinned_fit_object, from_container=False):
        """
        Construct an :py:obj:`UnbinnedPlotAdapter` for a :py:obj:`~kafe2.fit.unbinned.UnbinnedFit` object:
        :param unbinned_fit_object: an :py:obj:`~kafe2.fit.unbinned.UnbinnedFit` object
        :param from_container: Whether unbinned_fit_object was created ad-hoc from just a data
            container.
        :type from_container: bool
        """
        super(UnbinnedPlotAdapter, self).__init__(
            fit_object=unbinned_fit_object, from_container=from_container)
        self.n_plot_points = 100 if len(self.data_x) < 100 else len(self.data_x)

        self.x_range = add_pad_to_range(self._fit.data_range, scale=self.x_scale)
        _y = self.model_line_y
        _y_range = np.amin(_y), np.amax(_y)
        self.y_range = add_pad_to_range(_y_range, scale=self.y_scale)

    # -- public properties

    @property
    def data_x(self):
        """
        The 'x' coordinates of the data (used by :py:meth:`~plot_data`).

        :return: iterable
        """
        return self._fit.data

    @property
    def data_y(self):
        """
        The 'y' coordinates of the data (used by :py:meth:`~plot_data`).

        :return: iterable
        """
        raise TypeError("There's no y-data in the unbinned container")

    @property
    def data_xerr(self):
        """
        The magnitude of the data 'x' error bars (used by :py:meth:`~plot_data`).

        :return: iterable
        """
        return self._fit.data.err

    @property
    def data_yerr(self):
        raise TypeError("There's no y-data in the unbinned container, hence no y-error")

    @property
    def model_x(self):
        """x support values for model function"""
        return self.data_x

    @property
    def model_y(self):
        """
        The 'y' coordinates of the model (used by :py:meth:`~plot_model`).

        :return: iterable
        """
        return self._fit.model

    @property
    def model_line_x(self):
        """x support values for model function"""
        _xmin, _xmax = self.x_range
        if self.x_scale == 'linear':
            return np.linspace(_xmin, _xmax, self.n_plot_points)
        if self.x_scale == 'log':
            try:
                return np.geomspace(_xmin, _xmax, self.n_plot_points)
            except ValueError:
                raise ValueError("Support point calculation failed. The plot range can't include 0 "
                                 "when using log scale.")
        raise ValueError("x_range has to be one of {}. Found {} instead.".format(
            self.AVAILABLE_X_SCALES, self.x_scale))

    @property
    def model_line_y(self):
        """
        The 'y' coordinates of the model (used by :py:meth:`~plot_model`).

        :return: iterable
        """
        return self._fit.eval_model_function(x=self.model_line_x)

    @property
    def model_xerr(self):
        """
        The magnitude of the model 'x' error bars (used by :py:meth:`~plot_model`).

        :return: iterable
        """
        # Fixme: No static value
        return 0.5

    @property
    def model_yerr(self):
        """
        The magnitude of the model 'y' error bars (used by :py:meth:`~plot_model`).

        :return: iterable
        """
        return None

    @PlotAdapterBase.x_scale.setter
    def x_scale(self, scale):
        update_xrange = self.x_range == add_pad_to_range(self._fit.data_range, scale=self.x_scale)
        PlotAdapterBase.x_scale.fset(self, scale)  # use parent setter
        if update_xrange:
            self.x_range = add_pad_to_range(self._fit.data_range, scale=self.x_scale)

    # public methods
    def plot_data(self, target_axes, height=None, **kwargs):
        """
        Method called by the main plot routine to plot the data points to a specified matplotlib ``Axes`` object.

        :param target_axes: ``matplotlib`` ``Axes`` object
        :param height: The height of the lines which represent the density
        :type height: float
        :return: plot handle(s)
        """
        kwargs.pop('marker', None)  # pop marker keyword, as LineCollection doesn't support it

        if height is None:
            if self.y_scale == 'linear':
                height = self.y_range[1]/10  # set height to 1/10th of the max height of the model
            elif self.y_scale == 'log':
                height = 10**(np.log10(self.y_range[1])/10)

        data = self.data_x
        xy_pairs = np.column_stack([np.repeat(data, 2),
                                    np.tile([self.y_range[0], height], len(data))])
        lines = xy_pairs.reshape([len(data), 2, 2])
        line_segments = LineCollection(lines, **kwargs)
        return target_axes.add_collection(line_segments)

    def plot_model(self, target_axes, **kwargs):
        """
        Method called by the main plot routine to plot the model to a specified matplotlib ``Axes`` object.

        :param target_axes: ``matplotlib`` ``Axes`` object
        :param kwargs: keyword arguments accepted by the :py:func:`~kafe2._aux.step_fill_between` method
        :return: plot handle(s)
        """
        return target_axes.plot(self.model_x, self.model_y, **kwargs)

    def plot_ratio(self, target_axes, error_contributions=('data',), **kwargs):
        raise TypeError("Data/model ratio cannot be plotted for unbinned fits.")

    def plot_residual(self, target_axes, error_contributions=('data',), **kwargs):
        raise TypeError("Residuals cannot be plotted for unbinned fits.")

    def plot_model_line(self, target_axes, **kwargs):
        """
        Method called by the main plot routine to plot the model to a specified matplotlib ``Axes`` object.

        :param target_axes: ``matplotlib`` ``Axes`` object
        :param kwargs: keyword arguments accepted by the :py:func:`~kafe2._aux.step_fill_between` method
        :return: plot handle(s)
        """
        return target_axes.plot(self.model_line_x, self.model_line_y, **kwargs)



import abc
import six

from .._base import GenericDReprBase
# import fit classes
from ...custom import CustomFit
from ...histogram import HistFit
from ...indexed import IndexedFit
from ...xy import XYFit
from ...unbinned import UnbinnedFit

__all__ = ["FitDReprBase"]


@six.add_metaclass(abc.ABCMeta)
class FitDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = 'fit'

    _CLASS_TO_OBJECT_TYPE_NAME = {
        CustomFit: 'custom',
        HistFit: 'histogram',
        IndexedFit: 'indexed',
        UnbinnedFit: 'unbinned',
        XYFit: 'xy'
    }
    _OBJECT_TYPE_NAME_TO_CLASS = {
        'custom': CustomFit,
        'histogram': HistFit,
        'indexed': IndexedFit,
        'unbinned': UnbinnedFit,
        'xy': XYFit
    }

    def __init__(self, fit=None):
        self._kafe_object = fit
        super(FitDReprBase, self).__init__()



import sys
import warnings

from ...config import kc

__all__ = ['get_minimizer']

AVAILABLE_MINIMIZERS = dict()

_MINIMIZER_NAME_ALIASES = dict()

_catch_error_class = ImportError
if sys.version_info >= (3, 6):
    # python version 3.6+ throws a different exception type on import fail...
    _catch_error_class = ModuleNotFoundError

try:
    from .scipy_optimize_minimizer import MinimizerScipyOptimize
    __all__.append('MinimizerScipyOptimize')
    AVAILABLE_MINIMIZERS.update({
        'scipy': MinimizerScipyOptimize,
    })
    _MINIMIZER_NAME_ALIASES['scipy.optimize'] = 'scipy'
except _catch_error_class:
    pass

try:
    from .iminuit_minimizer import MinimizerIMinuit
    __all__.append('MinimizerIMinuit')
    AVAILABLE_MINIMIZERS.update({
        'iminuit': MinimizerIMinuit,
    })
except _catch_error_class:
    warnings.warn(
        "iminuit is not installed. While not strictly needed its use is "
        "heavily recommended for better performance and convergence."
    )
except SyntaxError:  # Newer versions of iminuit do not support Python 2.
    pass

try:
    from .root_tminuit_minimizer import MinimizerROOTTMinuit
    __all__.append('MinimizerROOTTMinuit')
    AVAILABLE_MINIMIZERS.update({
        'root.tminuit': MinimizerROOTTMinuit,
    })
    _MINIMIZER_NAME_ALIASES['minuit'] = 'root.tminuit'
    _MINIMIZER_NAME_ALIASES['root'] = 'root.tminuit'
except (_catch_error_class, ImportError):  # Python 2.7 ROOT bindings cause ImportError in Python 3.6
    pass

# raise if no minimizers can be imported
if not AVAILABLE_MINIMIZERS:
    raise RuntimeError("Fatal error: no minimizers found! Please ensure that "
                       "at least one of the following Python packages is installed: "
                       "['scipy', 'iminuit', 'ROOT']")


def get_minimizer(minimizer_spec=None):
    """Creates a MinimizerBase object from a given minimizer name.

    :param minimizer_spec: Name of the minimizer to return.
    :type minimizer_spec: str or None
    :return: MinimizerBase-derived kafe2 minimizer object
    """
    global AVAILABLE_MINIMIZERS
    # for 'None', return the default minimizer
    if minimizer_spec is None:
        # go through the default minimizers in the order specified in config
        _minimizer_specs = kc('core', 'minimizers', 'default_minimizer_list')

        # try every spec until a minimizer is found
        for _minimizer_spec in _minimizer_specs:
            _minimizer_spec = _minimizer_spec.lower()
            _minimizer_spec = _MINIMIZER_NAME_ALIASES.get(_minimizer_spec, _minimizer_spec)
            _minimizer = AVAILABLE_MINIMIZERS.get(_minimizer_spec, None)
            if _minimizer is not None:
                return _minimizer

        raise ValueError(
            "Could not find any minimizer in default list: {}! "
            "Available: {}".format(_minimizer_specs, list(AVAILABLE_MINIMIZERS.keys())))
    _minimizer_spec = minimizer_spec.lower()
    _minimizer_spec = _MINIMIZER_NAME_ALIASES.get(_minimizer_spec, _minimizer_spec)
    _minimizer = AVAILABLE_MINIMIZERS.get(_minimizer_spec, None)
    if _minimizer is not None:
        return _minimizer

    raise ValueError("Unknown minimizer '{}'! "
                     "Available: {}".format(minimizer_spec, list(AVAILABLE_MINIMIZERS.keys())))



import unittest
import sys
import numpy as np
import os
from kafe2 import XYFit, Plot


class TestXYPlot(unittest.TestCase):
    def setUp(self):
        self._ref_data = [[1, 2, 3], [0.9, 2.1, 3.0]]
        self._ref_dataset_label = 'My Dataset'
        self._ref_x_label = '$U$ [V]'
        self._ref_y_label = '$I$ [A]'
        self._ref_model_label = 'My Model'
        self._ref_error_label = self._ref_model_label + r' $\pm 1\sigma$'

        self.fit = XYFit(xy_data=self._ref_data)
        self.fit.add_error('y', 0.1)
        self.fit.do_fit()

        self.fit.data_container.label = self._ref_dataset_label
        self.fit.data_container.axis_labels = (self._ref_x_label, self._ref_y_label)
        self.fit.model_label = self._ref_model_label

        self.plot = Plot(self.fit)

    def test_warning_no_fit_performed(self):
        _fit = XYFit(xy_data=self._ref_data)
        _fit.add_error('y', 0.1)
        _plot = Plot(fit_objects=_fit)
        with self.assertWarns(Warning) as w:
            _plot.plot()
        self.assertIn("Did you forget to run fit.do_fit()?", str(w.warning))

    def test_plot_with_asymmetric_errors(self):
        self.plot.plot(asymmetric_parameter_errors=True)

    def test_labels_after_plotting(self):
        self.plot.plot()
        self.assertEqual(self.plot.axes[0]['main'].get_xlabel(), self._ref_x_label)
        self.assertEqual(self.plot.axes[0]['main'].get_ylabel(), self._ref_y_label)
        _, labels = self.plot.axes[0]['main'].get_legend_handles_labels()
        self.assertIn(self._ref_dataset_label, labels)
        self.assertIn(self._ref_model_label, labels)
        self.assertIn(self._ref_error_label, labels)

    def test_remove_labels(self):
        self.fit.data_container.label = '__del__'
        self.fit.data_container.axis_labels = ('__del__', '__del__')
        self.fit.model_label = '__del__'
        self.plot.plot()
        _, labels = self.plot.axes[0]['main'].get_legend_handles_labels()
        self.assertEqual(self.plot.axes[0]['main'].get_xlabel(), '')
        self.assertEqual(self.plot.axes[0]['main'].get_ylabel(), '')
        self.assertTrue(len(labels) == 0)

    def test_label_setter(self):
        test_x, test_y = 'Test_x', 'Test_y'
        self.plot.x_label = test_x
        self.plot.y_label = test_y
        self.plot.plot()
        self.assertEqual(self.plot.axes[0]['main'].get_xlabel(), test_x)
        self.assertEqual(self.plot.axes[0]['main'].get_ylabel(), test_y)

    def test_x_log_scale(self):
        x_range_lin = self.plot._get_plot_adapters()[0].x_range
        self.plot.x_scale = 'log'
        x_range_log = self.plot._get_plot_adapters()[0].x_range
        # check if additional pad has adapted to new scale
        self.assertNotEqual(x_range_lin, x_range_log)
        self.plot.x_range = (1e-3, 4)
        self.assertTrue(
            np.all(self.plot._get_plot_adapters()[0].model_line_x == np.geomspace(1e-3, 4, 200)))
        self.plot.plot()
        self.assertEqual(self.plot.axes[0]['main'].get_xscale(), 'log')

    def test_y_log_scale(self):
        self.plot.y_scale = 'log'
        self.plot.plot()
        self.assertEqual(self.plot.axes[0]['main'].get_yscale(), 'log')

    def test_no_ratio_no_residual(self):
        self.plot.plot(ratio=False, residual=False)
        _current_axes = self.plot._current_axes

        with self.subTest():
            self.assertIn('main', _current_axes)
        with self.subTest():
            self.assertNotIn('ratio', _current_axes)
        with self.subTest():
            self.assertNotIn('residual', _current_axes)

    def test_ratio_no_residual(self):
        self.plot.plot(ratio=True, residual=False)
        _current_axes = self.plot._current_axes

        with self.subTest():
            self.assertIn('main', _current_axes)
        with self.subTest():
            self.assertIn('ratio', _current_axes)
            _ylim = _current_axes['ratio'].get_ylim()
            self.assertAlmostEqual(1 - _ylim[0], _ylim[1] - 1)
            _model = self.fit.y_model
            _data = self.fit.y_data
            _err = self.fit.total_error
            self.assertAlmostEqual(
                _ylim[1],
                1 + 1.05 * np.max((_err + np.abs(_model - _data)) / np.abs(_model)),
            )
        with self.subTest():
            self.assertNotIn('residual', _current_axes)

    def test_no_ratio_residual(self):
        self.plot.plot(ratio=False, residual=True)
        _current_axes = self.plot._current_axes

        with self.subTest():
            self.assertIn('main', _current_axes)
        with self.subTest():
            self.assertNotIn('ratio', _current_axes)
        with self.subTest():
            self.assertIn('residual', _current_axes)
            _ylim = _current_axes['residual'].get_ylim()
            self.assertAlmostEqual(-_ylim[0], _ylim[1])
            _model = self.fit.y_model
            _data = self.fit.y_data
            _err = self.fit.total_error
            self.assertAlmostEqual(
                _ylim[1],
                1.05 * np.max(_err + np.abs(_model - _data)),
            )

    def test_ratio_residual(self):
        with self.assertRaises(NotImplementedError):
            self.plot.plot(ratio=True, residual=True)

    def test_save(self):
        self.plot.plot()

        self.plot.save()
        self.assertTrue(os.path.exists("fit.png"))
        os.remove("fit.png")

        self.plot.save("my_fit.png")
        self.assertTrue(os.path.exists("my_fit.png"))
        os.remove("my_fit.png")

    def test_get_keywords(self):
        self.plot.get_keywords("model")

        with self.assertRaises(ValueError):
            self.plot.get_keywords("This does not exist")


class TestMultiPlot(unittest.TestCase):
    def setUp(self):
        self._ref_data1 = [[1, 2, 3], [1, 2, 3]]
        self._ref_data2 = [[1.1, 1.9, 2.5], [5, 3, 2]]
        self.fit1 = XYFit(xy_data=self._ref_data1)
        self.fit2 = XYFit(xy_data=self._ref_data2)
        self.fit1.add_error('y', 0.1)
        self.fit1.do_fit()
        self.fit2.add_error('y', 0.1)
        self.fit2.do_fit()

        self.plot = Plot([self.fit1, self.fit2])
        self.plot_sep = Plot([self.fit1, self.fit2], separate_figures=True)

    def test_label_setter_str(self):
        x_label, y_label = "123x", "Y_test"
        for _p in (self.plot, self.plot_sep):
            _p.x_label = x_label
            _p.y_label = y_label
            _p.plot()
        self.assertEqual(self.plot.axes[0]['main'].get_xlabel(), x_label)
        self.assertEqual(self.plot.axes[0]['main'].get_ylabel(), y_label)
        for _ax in self.plot_sep.axes:
            self.assertEqual(_ax['main'].get_xlabel(), x_label)
            self.assertEqual(_ax['main'].get_ylabel(), y_label)

    def test_label_setter_list(self):
        x_labels = ("X1", "$x_2$")
        y_labels = ("123", "123")
        for _p in (self.plot, self.plot_sep):
            _p.x_label = x_labels
            _p.y_label = y_labels
            _p.plot()
        # x label should be concatenation
        self.assertEqual(self.plot.axes[0]['main'].get_xlabel(), ', '.join(x_labels))
        # y label should not contain the same label twice
        self.assertEqual(self.plot.axes[0]['main'].get_ylabel(), y_labels[0])
        # each axis should have different labels for separate plots
        for _ax, _x_label, _y_label in zip(self.plot_sep.axes, x_labels, y_labels):
            self.assertEqual(_ax['main'].get_xlabel(), _x_label)
            self.assertEqual(_ax['main'].get_ylabel(), _y_label)

    def test_range_setter(self):
        x_ranges = ([1, 5], (0, 6.12))
        y_ranges = [(0, 3), [1, 6]]
        self.plot_sep.x_range = x_ranges
        self.plot_sep.y_range = y_ranges
        for _adapter, _xr, _yr in zip(self.plot_sep._get_plot_adapters(), x_ranges, y_ranges):
            self.assertEqual(_adapter.x_range, tuple(_xr))
            self.assertEqual(_adapter.y_range, tuple(_yr))

    def test_scale_setter(self):
        self.plot.x_scale = 'log'
        self.plot.y_scale = 'log'
        self.plot.plot()
        self.assertEqual(self.plot.axes[0]['main'].get_xscale(), 'log')
        self.assertEqual(self.plot.axes[0]['main'].get_yscale(), 'log')

        x_scales = ('linear', 'log')
        y_scales = ('log', 'linear')
        self.plot_sep.x_scale = x_scales
        self.plot_sep.y_scale = y_scales
        for _ax, _xs, _ys in zip(self.plot_sep.axes, x_scales, y_scales):
            self.assertEqual(_ax['main'].get_xscale(), _xs)
            self.assertEqual(_ax['main'].get_yscale(), _ys)

    def test_save(self):
        self.plot.plot()
        self.plot_sep.plot()

        self.plot.save()
        self.assertTrue(os.path.exists("fit.png"))
        os.remove("fit.png")

        self.plot.save("my_fit.png")
        self.assertTrue(os.path.exists("my_fit.png"))
        os.remove("my_fit.png")

        self.plot_sep.save()
        self.assertTrue(os.path.exists("fit_0.png"))
        os.remove("fit_0.png")
        self.assertTrue(os.path.exists("fit_1.png"))
        os.remove("fit_1.png")

        self.plot_sep.save(figures=0)
        self.assertTrue(os.path.exists("fit_0.png"))
        os.remove("fit_0.png")
        self.plot_sep.save(figures=1)
        self.assertTrue(os.path.exists("fit_1.png"))
        os.remove("fit_1.png")

        self.plot_sep.save("my_fit.png")
        self.assertTrue(os.path.exists("my_fit_0.png"))
        os.remove("my_fit_0.png")
        self.assertTrue(os.path.exists("my_fit_1.png"))
        os.remove("my_fit_1.png")

        self.plot_sep.save("one_fit.png", figures=0)
        self.assertTrue(os.path.exists("one_fit.png"))
        os.remove("one_fit.png")
        self.plot_sep.save("another_fit", figures=1)
        self.assertTrue(os.path.exists("another_fit.png"))
        os.remove("another_fit.png")

        self.plot_sep.save(["one_fit.png", "another_fit.png"])
        self.assertTrue(os.path.exists("one_fit.png"))
        os.remove("one_fit.png")
        self.assertTrue(os.path.exists("another_fit.png"))
        os.remove("another_fit.png")

    def test_save_raise(self):
        self.plot_sep.plot()
        with self.assertRaises(ValueError):
            self.plot_sep.save(fname=1)
        with self.assertRaises(ValueError):
            self.plot_sep.save(fname=["fit_0.png", "fit_1.png", "fit_2.png"])
